/*
 * deagent_cmdu.c
 * File contains CMDUs handling needed for WiFi Data Elements.
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: saurabh.verma@iopsys.eu
 *	   anjan.chanda@iopsys.eu
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <net/if_arp.h>
#include <pthread.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <net/if.h>

#include <easy/easy.h>
#include <wifi.h>

#include "timer.h"
#include <cmdu.h>
#include <1905_tlvs.h>
#include <map2.h>
#include <cmdu_ackq.h>

#include "eventlog.h"
#include "wfa_delm.h"
#include "deagent.h"
#include "tlv.h"
#include "cmdu_validate.h"

#define UBUS_TIMEOUT            1000
#define FRAME_SIZE		4800
#define min(x, y)       (x) < (y) ? (x) : (y)




extern uint8_t *wifi_find_ie(uint8_t *ies, size_t len, uint8_t eid);
extern uint8_t *wifi_find_ie_ext(uint8_t *ies, size_t len, uint8_t ext_id);
extern void wifi_cap_set_from_ie(uint8_t *caps_bitmap, uint8_t *ie, size_t len);
extern void get_caps_info(enum wifi_band band, struct wifi_caps_element *caps,
		   struct wifi_caps *w_cap, uint8_t *bitmap);

static void deagent_send_cmdu_cb(struct ubus_request *req,
			      int type, struct blob_attr *msg)
{
	struct json_object *jobj = NULL;
	struct json_object *tmp;
	uint16_t *mid;
	char *str;

	if (!msg || !req->priv) {
		fprintf(stderr, "%s:Message recieved is NULL\n", __func__);
		return;
	}

	mid = (uint16_t *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	if (json_object_object_get_ex(jobj, "mid", &tmp))
		*mid = json_object_get_int(tmp);


	json_object_put(jobj);
}

int prepare_send_cmdu_blob(struct cmdu_buff *cmdu_data,
		struct blob_buf *b)
{
	char *blob_data = NULL;
	char dst_addr[18] = { 0 };

	blobmsg_add_u32(b, "type", cmdu_get_type(cmdu_data));
	blobmsg_add_u32(b, "mid", cmdu_get_mid(cmdu_data));

	hwaddr_ntoa(cmdu_data->origin, dst_addr);
	blobmsg_add_string(b, "dst", dst_addr);

	fprintf(stdout, "|%s:%d|cmdu:%02x|dst:%s|\n", __func__, __LINE__,
			cmdu_get_type(cmdu_data), dst_addr);

	if (cmdu_data->datalen) {
		char *tlv_str = NULL;
		uint16_t len = 0;

		len = (cmdu_data->datalen * 2) + 1;

		tlv_str = calloc(len, sizeof(char));
		if (!tlv_str)
			goto out;

		btostr(cmdu_data->data, cmdu_data->datalen, tlv_str);
		tlv_str[len-1] = '\0';

		blobmsg_add_string(b, "data", tlv_str);
		free(tlv_str);
	}

	blob_data = blobmsg_format_json(b->head, true);
	if (blob_data) {
		fprintf(stderr, "blob buf send data: |%s|\n", blob_data);
		free(blob_data);
	}

	return 0;
out:

	return -1;
}

static int deagent_send_cmdu_response(struct deagent_private *priv,
				   struct cmdu_buff *resp)
{
	uint16_t msgid = 0;
	int ret = 0;
	struct blob_buf b = { 0 };

	fprintf(stdout, "|%s:%d| Entry\n", __func__, __LINE__);

	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	ret = prepare_send_cmdu_blob(resp, &b);
	if (ret)
		goto out;

	ret = ubus_invoke(priv->ctx, priv->i1905, "cmdu",
			  b.head, deagent_send_cmdu_cb,
			  (void *)&msgid,
			  UBUS_TIMEOUT);
	if (ret) {
		fprintf(stderr, "[%s:%d] ubus call failed for |ieee1905 cmdu|",
				__func__, __LINE__);

		/* update the i1905 object */
		priv->i1905 = deagent_lookup_object(priv, IEEE1905_OBJECT);
		if (priv->i1905 == OBJECT_INVALID)
			fprintf(stderr, "Failed to get '%s' object!\n",
					IEEE1905_OBJECT);

		goto out;
	}

out:
	blob_buf_free(&b);

	return ret;
}

int deagent_send_cmdu_request(struct deagent_private *priv,
		struct cmdu_buff *req, uint16_t resp_type)
{
	int ret = 0;
	uint16_t msgid = 0;
	struct blob_buf b = { 0 };
	uint16_t message_type = cmdu_get_type(req);

	fprintf(stdout, "|%s:%d| Entry\n", __func__, __LINE__);

	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);
	ret = prepare_send_cmdu_blob(req, &b);
	if (ret)
		goto out;

	ret = ubus_invoke(priv->ctx, priv->i1905, "cmdu",
			  b.head, deagent_send_cmdu_cb,
			  (void *)&msgid,
			  UBUS_TIMEOUT);
	if (!ret) {
		fprintf(stderr, "%s: msg-id = %hu\n", __func__, msgid);

		/* Don't enqueue message id for topology notification,
		 * client disassociation stats & ap autoconfig search.
		 */
		if ((message_type == CMDU_TYPE_TOPOLOGY_NOTIFICATION) ||
				(message_type == CMDU_CLIENT_DISASSOCIATION_STATS) ||
				(message_type == CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH))
			goto out;

		/* Store mid of the sent CMDU with an expected response type.
		 * On receiving a 'resp_type' cmdu, lookup in the ackq and
		 * drop or handle it.
		 */
		//cmdu_ackq_enqueue(&priv->cmdu_ack_q, resp_type, msgid,
		//		  req->origin, 2000);  //FIXME-CR
	} else {
		fprintf(stderr, "[%s:%d] ubus call failed for |ieee1905 cmdu|",
				__func__, __LINE__);

		/* unset getter_running,
		 * if not set, refresh method won't work
		 */
		priv->getter_running = 0;

		/* update the i1905 object */
		priv->i1905 = deagent_lookup_object(priv, IEEE1905_OBJECT);
		if (priv->i1905 == OBJECT_INVALID)
			fprintf(stderr, "Failed to get '%s' object!\n",
					IEEE1905_OBJECT);
	}

out:
	blob_buf_free(&b);

	return ret;
}

static int get_sta_index(struct wifi_bss_element *bss, uint8_t *sta)
{
	int i = 0;
	char addr[18] = {0};

	for (i = 0; i < bss->num_stations; i++) {
		if (hwaddr_equal(bss->stalist[i].macaddr, sta)) {
			hwaddr_ntoa(sta, addr);
			fprintf(stdout, "found sta:|%s| at index %d\n", addr, i);
			return i;
		}
	}
	return -1;
}

static int get_opclasss_index(struct wifi_radio_element *radio, uint8_t opclass_id)
{
	int i = 0;
	struct wifi_scanres_opclass_element *opclass = NULL;

	/* TODO: Assuming single scanresult per radio */
	if (radio->scanlist) {
		for (i = 0; i < radio->scanlist[0].num_opclass_scanned; i++) {
			opclass = radio->scanlist[0].opclass_scanlist + i;
			if (opclass->opclass == opclass_id)
				return i;
		}
	}

	return -1;
}

static int get_channel_index(struct wifi_scanres_opclass_element *opclass, uint8_t channel_id)
{
	int i = 0;
	struct wifi_scanres_channel_element *ch = NULL;

	for (i = 0; i < opclass->num_channels_scanned; i++) {
		ch = opclass->channel_scanlist + i;
		if (ch->channel == channel_id)
			return i;
	}

	return -1;
}

static int get_radio_index(struct wifi_network_device *dev, uint8_t *radio_id)
{
	int i = 0;
	char addr[18] = {0};

	for (i = 0; i < dev->num_radios; i++) {
		if (hwaddr_equal(dev->radiolist[i].macaddr, radio_id)) {
			hwaddr_ntoa(radio_id, addr);
			fprintf(stdout, "found radio:|%s| at index %d\n", addr, i);
			return i;
		}
	}
	return -1;
}

static int get_bss_index(struct wifi_radio_element *radio, uint8_t *bssid)
{
	int i = 0;
	char addr[18] = {0};

	for (i = 0; i < radio->num_bss; i++) {
		if (hwaddr_equal(radio->bsslist[i].bssid, bssid)) {
			hwaddr_ntoa(bssid, addr);
			fprintf(stdout, "found bss:|%s| at index %d\n", addr, i);
			return i;
		}
	}
	return -1;
}

static int get_radio_bss_sta_index(struct wifi_network_device *dev, uint8_t *sta,
		int *radio_index, int *bss_index)
{
	int i, j;
	int sta_index;

	for (i = 0; i < dev->num_radios; i++) {
		for (j = 0; j < dev->radiolist[i].num_bss; j++) {
			sta_index = get_sta_index(&dev->radiolist[i].bsslist[j], sta);
			if (sta_index != -1) {
				*radio_index = i;
				*bss_index = j;
				return sta_index;
			}
		}
	}

	return -1;
}

static int get_radio_and_bss_index(struct wifi_network_device *dev,
		uint8_t *bssid, int *radio_index)
{
	int i;
	int bss_index;

	for (i = 0; i < dev->num_radios; i++) {
		bss_index = get_bss_index(&dev->radiolist[i], bssid);
		if (bss_index != -1) {
			*radio_index = i;
			return bss_index;
		}
	}

	return -1;
}

static struct wifi_assoc_frame *wifi_get_frame(struct deagent_private *p,
		uint8_t *macaddr)
{
	struct wifi_assoc_frame *fr = NULL;

	list_for_each_entry(fr, &p->framelist, list) {
		fprintf(stderr, "frame mac: " MACFMT " input: " MACFMT "\n",
				MAC2STR(fr->macaddr),
				MAC2STR(macaddr));
		if (hwaddr_equal(fr->macaddr, macaddr))
			return fr;
	}

	return NULL;
}

void fill_assoc_sta_caps(struct tlv_client_cap_report *p, int len,
	struct wifi_caps_element *caps)
{
	uint8_t *ht_ie, *vht_ie, *he_ie;
	uint8_t cbitmap[16] = {0};
	struct wifi_caps w_caps;

	if (!p->frame)
		return;

	memset(&w_caps, 0, sizeof(struct wifi_caps));
	ht_ie = wifi_find_ie(p->frame, len, IE_HT_CAP);
	vht_ie = wifi_find_ie(p->frame, len, IE_VHT_CAP);
	he_ie = wifi_find_ie_ext(p->frame, len, IE_EXT_HE_CAP);

	if (ht_ie) {
		fprintf(stderr, "HT caps...\n");
		w_caps.valid |= WIFI_CAP_HT_VALID;
		memcpy(&w_caps.ht, &ht_ie[2], ht_ie[1]);
		wifi_cap_set_from_ie(cbitmap, ht_ie, ht_ie[1] + 2);
	}

	if (vht_ie) {
		fprintf(stderr, "VHT caps...\n");
		w_caps.valid |= WIFI_CAP_VHT_VALID;
		memcpy(&w_caps.vht, &vht_ie[2], vht_ie[1]);
		wifi_cap_set_from_ie(cbitmap, vht_ie, vht_ie[1] + 2);
	}

	if (he_ie) {
		fprintf(stderr, "HE caps...\n");
		w_caps.valid |= WIFI_CAP_HE_VALID;
		memcpy(&w_caps.he, &he_ie[3], min(he_ie[1], sizeof(struct wifi_caps_he)));
		wifi_cap_set_from_ie(cbitmap, he_ie, he_ie[1] + 2);
	}

	/* TODO/FIXME:
	 * pass BAND_5 as first argument,
	 * need to fetch this info from raw frame
	 * need band info for HE capability.
	 */
	get_caps_info(BAND_5, caps, &w_caps, cbitmap);
}

int deagent_handle_autoconfig_response(struct deagent_private *p,
				    struct cmdu_buff *resp)
{
	bool is_cntlr = false;
	int i;
	struct tlv *tv[4][16];

	if (!validate_autoconfig_response(resp, tv)) {
		fprintf(stderr, "%s: cmdu validation "\
				"[AP_AUTOCONFIGURATION_RESPONSE] failed\n",
				__func__);
		return -1;
	}

	if (tv[0][0]->data[0] > 0) {
		for (i = 0; i < tv[0][0]->data[0]; i++) {
			if (tv[0][0]->data[i+1] ==
					SUPPORTED_SERVICE_MULTIAP_CONTROLLER) {
				is_cntlr = true;
				break;
			}
		}
	}

	if (!is_cntlr) {
		fprintf(stderr, "Autoconfig response was not from controller\n");
		return -1;
	}

	fprintf(stdout, "controller id: " MACFMT "\n", MAC2STR(resp->origin));

	/* store controller id */
	memcpy(p->cntlr_id, resp->origin, 6);

	return 0;
}

static int prepare_topology_response(struct deagent_private *priv,
		struct cmdu_buff *rec_cmdu)
{
	int ret;
	uint16_t mid = 0;
	struct cmdu_buff *resp;
	struct wifi_network_device *own = NULL;
	struct tlv *tv[1][16] = {0};

	if (!validate_topology_query(rec_cmdu, tv)) {
		fprintf(stderr, "%s: cmdu validation: "\
				"[TOPOLOGY_QUERY] failed\n", __func__);
		return -1;
	}

	/* for profile-2, check multi-AP TLV
	 */
	if ((priv->cfg.profile == MAP_PROFILE_2) && (!tv[0][0]))
			return -1;

	if (priv->dm_available)
		own = &priv->dm->network.device[0];

	if (!own)
		return -1;

	mid = cmdu_get_mid(rec_cmdu);
	resp = cmdu_alloc_simple(CMDU_TYPE_TOPOLOGY_RESPONSE, &mid);
	if (!resp) {
		fprintf(stdout, "%s: -ENOMEM\n", __func__);
		return -1;
	}

	memcpy(resp->origin, rec_cmdu->origin, 6);

	/* self node ieee1905 specific tlvs:
	 * device information type TLV
	 * device bridging capability TLV
	 * non-1905 neighbor device list TLV
	 * 1905.1 neighbor device TLV
	 */
	ret = deagent_gen_i1905_tlvs(own, resp);
	if (ret)
		goto error;

	/* SupportedService TLV */
	ret = deagent_gen_supported_service(priv, resp,
			SUPPORTED_SERVICE_MULTIAP_AGENT);
	if (ret)
		goto error;

	/* AP Operational BSS TLV */
	ret = deagent_gen_ap_oper_bss(own, resp);
	if (ret)
		goto error;

	/* Associated Client TLV */
	ret = deagent_gen_assoc_client(own, resp);
	if (ret)
		goto error;

	/* vendor specific tlv
	 * not a part of topology response
	 * need to add to align the response
	 * with map-agent (map-topology specific)
	 * this having sta stats only.
	 */
	ret = deagent_gen_vendor_specific(own, resp);
	if (ret)
		goto error;

	if (priv->cfg.profile == MAP_PROFILE_2) {
		/* Multi-AP profile */
		ret = deagent_gen_map_profile(priv, resp,
				MULTIAP_PROFILE_2);
		if (ret)
			goto error;
	}

	cmdu_put_eom(resp);

	ret = deagent_send_cmdu_response(priv, resp);
	cmdu_free(resp);
	return 0;

error:
	cmdu_free(resp);
	return -1;
}

static int prepare_ap_capability_response(struct deagent_private *priv,
					struct cmdu_buff *rec_cmdu)
{
	int i, ret;
	uint16_t mid = 0;
	struct cmdu_buff *resp;
	struct wifi_network_device *own = NULL;

	if (priv->dm_available)
		own = &priv->dm->network.device[0];

	if (!own)
		return -1;

	mid = cmdu_get_mid(rec_cmdu);
	resp = cmdu_alloc_simple(CMDU_AP_CAPABILITY_REPORT, &mid);
	if (!resp) {
		fprintf(stderr, "%s: -ENOMEM\n", __func__);
		return -1;
	}

	memcpy(resp->origin, rec_cmdu->origin, 6);

	/* AP Capability TLV */
	ret = deagent_gen_ap_cap(own, resp);
	if (ret)
		goto error;

	/* AP Radio Basic Capabilities TLV */
	for (i = 0; i < own->num_radios; i++) {
		ret = deagent_gen_ap_radio_basic_cap(own, resp, i);
		if (ret)
			goto error;
	}

	/* AP HT Capabilities TLV */
	for (i = 0; i < own->num_radios; i++) {
		if (!(own->radiolist[i].caps.valid & WIFI_CAP_HT_VALID))
			continue;

		ret = deagent_gen_ap_ht_cap(own, resp, i);
		if (ret)
			goto error;
	}

	/* AP VHT Capabilities TLV */
	for (i = 0; i < own->num_radios; i++) {
		if (!(own->radiolist[i].caps.valid & WIFI_CAP_VHT_VALID))
			continue;

		ret = deagent_gen_ap_vht_cap(own, resp, i);
		if (ret)
			goto error;
	}

	/* AP HE Capabilities TLV */
	for (i = 0; i < own->num_radios; i++) {
		if (!(own->radiolist[i].caps.valid & WIFI_CAP_HE_VALID))
			continue;

		ret = deagent_gen_ap_he_cap(own, resp, i);
		if (ret)
			goto error;
	}

	if (priv->cfg.profile == MAP_PROFILE_2) {
		/* Channel Scan Capabilities TLV */
		ret = deagent_gen_channel_scan_capability(own, resp);
		if (ret)
			goto error;

		/* CAC Capabilities TLV */
		ret = deagent_gen_cac_cap(own, resp);
		if (ret)
			goto error;

		/* Profile-2 AP Capability TLV */
		ret = deagent_gen_profile2_ap_cap(own, resp);
		if (ret)
			goto error;

		/* Metric Collection Interval TLV */
		ret = deagent_gen_metric_collection_int(own, resp);
		if (ret)
			goto error;
	}

	cmdu_put_eom(resp);

	ret = deagent_send_cmdu_response(priv, resp);
	cmdu_free(resp);

	return 0;

error:
	cmdu_free(resp);

	return -1;
}

static int prepare_apstats_response(struct deagent_private *priv,
					struct cmdu_buff *rec_cmdu)
{
	int i = 0, j = 0, k;
	int ret;
	uint16_t mid = 0;
	int radio_index;
	int bss_index;
	struct cmdu_buff *resp;
	struct wifi_network_device *own = NULL;
	struct wifi_radio_element *radio;
	struct wifi_bss_element *bss;
	struct tlv *tv[2][16] = {0};

	if (!validate_ap_metric_query(rec_cmdu, tv)) {
		fprintf(stderr, "%s: cmdu validation: " \
				"[AP_METRIC_QUERY] failed\n", __func__);
		return -1;
	}

	if (priv->dm_available)
		own = &priv->dm->network.device[0];

	if (!own)
		return -1;

	mid = cmdu_get_mid(rec_cmdu);

	resp = cmdu_alloc_simple(CMDU_AP_METRICS_RESPONSE, &mid);
	if (!resp) {
		fprintf(stderr, "%s: -ENOMEM\n", __func__);
		return -1;
	}

	memcpy(resp->origin, rec_cmdu->origin, 6);

	/* AP Radio Identifier TLV */
	while (tv[1][i]) {
		struct tlv_ap_radio_identifier *tmp =
			(struct tlv_ap_radio_identifier *)tv[1][i++]->data;

		radio_index = get_radio_index(own, tmp->radio);
		if ((radio_index != -1) &&
				(priv->cfg.profile == MAP_PROFILE_2)) {
			/* Radio Metrics TLV */
			ret = deagent_gen_radio_metrics(own, resp, radio_index);
			if (ret)
				goto error;
		}
	}

	/* AP Metric Query TLV */
	if (tv[0][0]) {
		struct tlv_ap_metric_query *tmp =
			(struct tlv_ap_metric_query *)tv[0][0]->data;

		for (j = 0; j < tmp->num_bss; j++) {
			bss_index = get_radio_and_bss_index(own,
					tmp->bss[j].bssid,
					&radio_index);

			if (bss_index == -1)
				continue;

			radio = own->radiolist + radio_index;
			bss = radio->bsslist + bss_index;
			/* AP Metrics TLV */
			ret = deagent_gen_ap_metrics(own, resp,
					radio_index, bss_index);
			if (ret)
				goto error;

			if (priv->cfg.profile == MAP_PROFILE_2) {
				/* AP Extended Metrics TLV */
				ret = deagent_gen_ap_ext_metrics(own, resp,
						radio_index, bss_index);
				if (ret)
					goto error;
			}

			for (k = 0; k < bss->num_stations; k++) {
				/* Associated STA Traffic Stats TLV */
				ret = deagent_gen_assoc_sta_traffic_stats(own, resp,
						radio_index, bss_index, k);
				if (ret)
					goto error;

				/* Associated STA Link Metrics TLV */
				ret = deagent_gen_assoc_sta_link_metrics(own, resp,
						radio_index, bss_index, k);
				if (ret)
					goto error;

				if (priv->cfg.profile == MAP_PROFILE_2) {
					/* Associated STA Extended Link Metrics TLV */
					ret = deagent_gen_assoc_sta_ext_link_metric(
							own, resp, radio_index,
							bss_index, k);
					if (ret)
						goto error;
				}

			}

		}
	}

	cmdu_put_eom(resp);

	ret = deagent_send_cmdu_response(priv, resp);
	cmdu_free(resp);

	return 0;

error:
	cmdu_free(resp);

	return -1;
}

static int prepare_backhaul_response(struct deagent_private *priv,
					struct cmdu_buff *rec_cmdu)
{
	int i, ret;
	uint16_t mid = 0;
	struct cmdu_buff *resp;
	struct wifi_network_device *own = NULL;

	if (priv->dm_available)
		own = &priv->dm->network.device[0];

	if (!own)
		return -1;

	mid = cmdu_get_mid(rec_cmdu);
	resp = cmdu_alloc_simple(CMDU_BACKHAUL_STA_CAPABILITY_REPORT, &mid);
	if (!resp) {
		fprintf(stderr, "%s: -ENOMEM\n", __func__);
		return -1;
	}

	memcpy(resp->origin, rec_cmdu->origin, 6);

	for (i = 0; i < own->num_radios; i++) {
		ret = deagent_gen_bsta_radio_cap(own, resp, i);
		if (ret)
			goto error;
	}

	cmdu_put_eom(resp);
	ret = deagent_send_cmdu_response(priv, resp);
	cmdu_free(resp);

	return 0;

error:
	cmdu_free(resp);

	return -1;
}

static int prepare_unassoc_sta_response(struct deagent_private *priv,
					struct cmdu_buff *rec_cmdu)
{
	/* TODO:FIXME: Not supported
	 * Currently set to the 0/NULL from 17.2.6
	 */
	return 0;
}

static int prepare_operating_channel_response(struct deagent_private *priv,
						struct cmdu_buff *rec_cmdu)
{
	int i, ret;
	uint16_t mid = 0;
	struct cmdu_buff *resp;
	struct wifi_network_device *own = NULL;

	if (priv->dm_available)
		own = &priv->dm->network.device[0];

	if (!own)
		return -1;

	mid = cmdu_get_mid(rec_cmdu);
	resp = cmdu_alloc_simple(CMDU_OPERATING_CHANNEL_REPORT, &mid);
	if (!resp) {
		fprintf(stderr, "%s: -ENOMEM\n", __func__);
		return -1;
	}

	memcpy(resp->origin, rec_cmdu->origin, 6);

	for (i = 0; i < own->num_radios; i++) {
		/* Operating Channel Report TLV */
		ret = deagent_gen_oper_channel_report(own, resp, i);
		if (ret)
			goto error;
	}

	cmdu_put_eom(resp);
	ret = deagent_send_cmdu_response(priv, resp);
	cmdu_free(resp);

	return 0;
error:
	cmdu_free(resp);

	return -1;
}

static int prepare_scanres_channel_response(struct deagent_private *priv,
						struct cmdu_buff *rec_cmdu)
{
	int i, j, k;
	int ret;
	int offset = 0;
	int radio_index;
	int opclass_index;
	int channel_index;
	uint16_t mid = 0;
	struct cmdu_buff *resp;
	struct wifi_network_device *own = NULL;
	struct wifi_radio_element *radio;
	struct wifi_scanres_opclass_element *opclass;
	struct tlv *tv[1][16];

	if (!validate_channel_scan_request(rec_cmdu, tv)) {
		fprintf(stderr, "%s: cmdu validation: " \
				"[CHANNEL_SCAN_REQ] failed\n", __func__);
		return -1;
	}

	if (priv->dm_available)
		own = &priv->dm->network.device[0];

	if (!own)
		return -1;

	mid = cmdu_get_mid(rec_cmdu);
	resp = cmdu_alloc_frame(FRAME_SIZE);
	if (!resp) {
		fprintf(stderr, "%s: -ENOMEM\n", __func__);
		return -1;
	}

	cmdu_set_mid(resp, mid);
	cmdu_set_type(resp, CMDU_CHANNEL_SCAN_REPORT);
	CMDU_SET_LAST_FRAGMENT(resp->cdata);
	memcpy(resp->origin, rec_cmdu->origin, 6);

	if (tv[0][0]) {
		uint8_t *q_data = (uint8_t *)tv[0][0]->data;
		struct tlv_channel_scan_request *tmp =
			(struct tlv_channel_scan_request *)q_data;

		/* Timestamp TLV */
		ret = deagent_gen_timestamp(own, resp);
		if (ret)
			goto error;

		offset = 1 + 1;
		for (i = 0; i < tmp->num_radio; i++) {
			struct channel_scan_request_radio *r =
				(struct channel_scan_request_radio *)&q_data[offset];

			offset += sizeof(*r);
			radio_index = get_radio_index(own, r->radio);
			if (radio_index == -1)
				continue;

			radio = own->radiolist + radio_index;
			for (j = 0; j < r->num_opclass; j++) {
				struct channel_scan_request_opclass *op =
					(struct channel_scan_request_opclass *)&q_data[offset];

				offset += sizeof(*op) + op->num_channel;
				opclass_index = get_opclasss_index(radio, op->classid);
				if (opclass_index == -1)
					continue;

				opclass = radio->scanlist[0].opclass_scanlist + opclass_index;
				if (op->num_channel > 0) {
					for (k = 0; k < op->num_channel; k++) {
						channel_index = get_channel_index(opclass, op->channel[k]);
						if (channel_index == -1)
							continue;

						/* Channel Scan Result TLV */
						ret = deagent_gen_channel_scan_result(own, resp,
								radio_index, opclass_index, channel_index);
						if (ret)
							goto error;
					}
				} else {
					for (k = 0; k < opclass->num_channels_scanned; k++) {
						/* Channel Scan Result TLV */
						ret = deagent_gen_channel_scan_result(own, resp,
								radio_index, opclass_index, k);
						if (ret)
							goto error;
					}

				}
			}
		}
	}

	memcpy(resp->origin, rec_cmdu->origin, 6);

	cmdu_put_eom(resp);
	ret = deagent_send_cmdu_response(priv, resp);
	cmdu_free(resp);

	return 0;
error:
	cmdu_free(resp);

	return -1;
}

static int prepare_client_capability_response(struct deagent_private *priv,
		struct cmdu_buff *rec_cmdu)
{
	bool is_sta_pres = false;
	uint8_t result_code = 0x00;
	uint8_t reason_code = 0x00;
	uint16_t mid = 0;
	int radio_index, bss_index;
	int ret, sta_index = -1;
	struct cmdu_buff *resp;
	struct wifi_assoc_frame *assoc_frame;
	struct wifi_network_device *own = NULL;
	struct tlv_client_info *tmp;
	struct tlv *tv[1][16];

	if (!validate_client_cap_query(rec_cmdu, tv)) {
		fprintf(stderr, "%s: cmdu_validation "\
				"[CLIENT_CAPABILITY_QUERY] failed\n", __func__);
		return -1;
	}

	tmp = (struct tlv_client_info *)tv[0][0]->data;

	if (priv->dm_available)
		own = &priv->dm->network.device[0];

	if (own) {
		sta_index = get_radio_bss_sta_index(own, tmp->macaddr,
				&radio_index, &bss_index);
		if (sta_index != -1)
			is_sta_pres = true;
	}

	mid = cmdu_get_mid(rec_cmdu);
	resp = cmdu_alloc_simple(CMDU_CLIENT_CAPABILITY_REPORT, &mid);
	if (!resp) {
		fprintf(stderr, "%s: -ENOMEM\n", __func__);
		return -1;
	}

	memcpy(resp->origin, rec_cmdu->origin, 6);


	/* Client Info TLV */
	ret = deagent_gen_client_info(priv, resp, tmp->macaddr, tmp->bssid);
	if (ret)
		goto error;

	assoc_frame = wifi_get_frame(priv, tmp->macaddr);
	if (assoc_frame && is_sta_pres) {
		/* sta is associated */
		fprintf(stderr, "sta associated\n");
		/* Client Capability Report TLV */
		ret = deagent_gen_client_cap_report(priv, resp,
				result_code, assoc_frame);
		if (ret)
			goto error;

	} else {
		result_code = 0x01;

		if (is_sta_pres) {
			/* assoc frame is not available */
			fprintf(stderr, "no assoc frame for sta " MACFMT "\n",
					MAC2STR(tmp->macaddr));
			reason_code = 0x03;
		} else {
			/* sta is not associated with any BSS*/
			fprintf(stderr, "sta not associated\n");
			reason_code = 0x02;
		}

		/* Client Capability Report TLV */
		ret = deagent_gen_client_cap_report(priv, resp,
				result_code, NULL);
		if (ret)
			goto error;

		/* Error Code TLV */
		ret = deagent_gen_error_code(priv, resp,
				reason_code, tmp->macaddr);
		if (ret)
			goto error;

	}

	cmdu_put_eom(resp);
	deagent_send_cmdu_response(priv, resp);
	cmdu_free(resp);

	return 0;
error:
	cmdu_free(resp);

	return -1;
}

static int prepare_topology_notication(struct deagent_private *priv, uint8_t *sta_mac,
					const char *event, uint8_t *bssid)
{
	int ret;
	uint16_t mid = 0;
	uint16_t resp_expected = 0x00;
	struct cmdu_buff *resp;

	resp = cmdu_alloc_simple(CMDU_TYPE_TOPOLOGY_NOTIFICATION, &mid);
	if (!resp) {
		fprintf(stderr, "%s: -ENOMEM\n", __func__);
		return -1;
	}

	CMDU_SET_RELAY_MCAST(resp->cdata);
	memcpy(resp->origin, "\x01\x80\xC2\x00\x00\x13", 6);

	/* 1905.1 AL MAC address type TLV */
	ret = deagent_gen_aladdr(priv, resp);
	if (ret)
		goto error;

	/* Client Association Event TLV */
	ret = deagent_gen_client_assoc_event(priv, resp, sta_mac, event, bssid);
	if (ret)
		goto error;

	cmdu_put_eom(resp);
	ret = deagent_send_cmdu_request(priv, resp, resp_expected);
	cmdu_free(resp);

	return 0;

error:
	cmdu_free(resp);

	return -1;
}

static int prepare_disassoc_stats(struct deagent_private *priv,
		uint8_t *sta_mac, int reason_code)
{
	bool is_sta = false;
	int i, ret;
	uint16_t mid = 0;
	uint16_t resp_expected = 0x00;
	struct cmdu_buff *resp;
	struct wifi_sta_events *sv = priv->sv;
	struct wifi_disassoc_events *disassoc_ev = NULL;

	for (i = (sv->num_disassoc_events - 1); i >= 0; i--) {
		disassoc_ev = sv->disassoclist + i;
		if (hwaddr_equal(disassoc_ev->sta_data.macaddr, sta_mac)) {
			is_sta = true;
			break;
		}
	}

	if (!is_sta)
		return -1;

	resp = cmdu_alloc_simple(CMDU_CLIENT_DISASSOCIATION_STATS, &mid);
	if (!resp) {
		fprintf(stderr, "%s: -ENOMEM\n", __func__);
		return -1;
	}

	memcpy(resp->origin, priv->cntlr_id, 6);

	/* STA MAC Address TLV */
	ret = deagent_gen_sta_mac(priv, resp, sta_mac);
	if (ret)
		goto error;

	/* Reason Code TLV */
	ret = deagent_gen_reason_code(priv, resp, reason_code);
	if (ret)
		goto error;

	/* Associated STA Traffic Stats TLV */
	ret = deagent_gen_disassoc_stats(priv, resp, sta_mac, disassoc_ev);
	if (ret)
		goto error;

	cmdu_put_eom(resp);
	ret = deagent_send_cmdu_request(priv, resp, resp_expected);
	cmdu_free(resp);

	return 0;
error:
	cmdu_free(resp);

	return -1;
}

int prepare_autoconfig_search(struct deagent_private *priv, uint8_t band)
{
	int ret;
	uint16_t mid = 0;
	uint16_t resp_expected;
	struct cmdu_buff *req;

	req = cmdu_alloc_simple(CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH, &mid);
	if (!req) {
		fprintf(stderr, "%s: -ENOMEM\n", __func__);
		return -1;
	}

	CMDU_SET_RELAY_MCAST(req->cdata);
	memcpy(req->origin, "\x01\x80\xC2\x00\x00\x13", 6);
	resp_expected = CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE;

	/* 1905.1 AL MAC address type TLV */
	ret = deagent_gen_aladdr(priv, req);
	if (ret)
		goto error;

	/* SearchedRole TLV */
	ret = deagent_gen_searched_role(priv, req,
			IEEE80211_ROLE_REGISTRAR);
	if (ret)
		goto error;

	/* AutoconfigFreqBand TLV */
	ret = deagent_gen_autoconfig_band(priv, req, band);
	if (ret)
		goto error;

	/* SupportedService TLV */
	ret = deagent_gen_supported_service(priv, req,
			SUPPORTED_SERVICE_MULTIAP_AGENT);
	if (ret)
		goto error;

	/* SearchedService TLV */
	ret = deagent_gen_searched_service(priv, req,
			SEARCHED_SERVICE_MULTIAP_CONTROLLER);
	if (ret)
		goto error;

	if (priv->cfg.profile == MAP_PROFILE_2) {
		/* Multi-AP Profile TLV */
		ret = deagent_gen_map_profile(priv, req,
				MULTIAP_PROFILE_2);
		if (ret)
			goto error;
	}

	deagent_send_cmdu_request(priv, req, resp_expected);
	cmdu_free(req);

	return 0;
error:
	cmdu_free(req);

	return -1;
}

int deagent_handle_map_request(struct deagent_private *p, struct cmdu_buff *cmdu)
{
	int ret = 0;
	uint16_t message_type = cmdu_get_type(cmdu);

	switch (message_type) {
	case CMDU_TYPE_TOPOLOGY_QUERY:
		ret = prepare_topology_response(p, cmdu);
		break;
	case CMDU_AP_CAPABILITY_QUERY:
		ret = prepare_ap_capability_response(p, cmdu);
		break;
	case CMDU_AP_METRICS_QUERY:
		ret = prepare_apstats_response(p, cmdu);
		break;
	case CMDU_BACKHAUL_STA_CAPABILITY_QUERY:
		ret = prepare_backhaul_response(p, cmdu);
		break;
	case CMDU_UNASSOC_STA_LINK_METRIC_QUERY:
		ret = prepare_unassoc_sta_response(p, cmdu);
		break;
	case CMDU_CHANNEL_SCAN_REQUEST:
		ret = prepare_scanres_channel_response(p, cmdu);
		break;
	case CMDU_CHANNEL_SELECTION_REQ:
		ret = prepare_operating_channel_response(p, cmdu);
		break;
	case CMDU_CLIENT_CAPABILITY_QUERY:
		ret = prepare_client_capability_response(p, cmdu);
		break;
	default:
		break;
	}

	return ret;
}


int deagent_handle_map_event(struct deagent_private *p, struct cmdu_buff *cmdu)
{
	int ret = 0;
	uint16_t message_type = cmdu_get_type(cmdu) & 0xffff;

	switch (message_type) {
	case CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE:
		fprintf(stderr, "%s: priv = %p (%s) cmdu = %02x\n",
			__func__, p, "agent", message_type);

		ret = deagent_handle_autoconfig_response(p, cmdu);
		break;

	case CMDU_TYPE_TOPOLOGY_QUERY:
	case CMDU_AP_CAPABILITY_QUERY:
	case CMDU_AP_METRICS_QUERY:
	case CMDU_CLIENT_CAPABILITY_QUERY:
	case CMDU_BACKHAUL_STA_CAPABILITY_QUERY:
	case CMDU_UNASSOC_STA_LINK_METRIC_QUERY:
	case CMDU_CHANNEL_SCAN_REQUEST:
	case CMDU_CHANNEL_SELECTION_REQ:
		fprintf(stderr, "%s: priv = %p cmdu = %02x\n",
			__func__, p, message_type);
		ret = deagent_handle_map_request(p, cmdu);
		break;
	default:
		break;
	}

	return ret;
}

static void check_ap_stalist_cb(struct ubus_request *req,
				int mtype, struct blob_attr *msg)
{
	int sta_count = 0;
	int i;
	char *str = NULL;
	char *addr = NULL;
	struct json_object *jobj = NULL;
	struct json_object *assoc_arr, *assoc_obj, *mac;
	uint8_t tmp_mac[6] = {0};
	struct sta_query {
		uint8_t sta_mac[6];
		bool is_present;
	} *tmp = NULL;

	if (!msg || !req->priv) {
		fprintf(stderr, "%s: msg = NULL\n", __func__);
		return;
	}

	tmp = (struct sta_query *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	/* Check the AP assoclist */
	if (json_object_object_get_ex(jobj, "assoclist", &assoc_arr)) {
		sta_count = json_object_array_length(assoc_arr);
		for (i = 0; i < sta_count; i++) {
			assoc_obj = json_object_array_get_idx(assoc_arr, i);
			if (json_object_object_get_ex(assoc_obj, "macaddr", &mac)) {
				addr = strdup(json_object_get_string(mac));
				hwaddr_aton(addr, tmp_mac);
				free(addr);
				if (hwaddr_equal(tmp_mac, tmp->sta_mac)) {
					tmp->is_present = true;
					break;
				}
			}
		}
	}

	json_object_put(jobj);
}

static void get_associated_bss_cb(struct ubus_request *req,
				int mtype, struct blob_attr *msg)
{
	char *str = NULL;
	struct json_object *jobj = NULL, *tmp;
	uint8_t *bssid;
	char *addr = NULL;

	if (!msg || !req->priv) {
		fprintf(stderr, "%s: msg = NULL\n", __func__);
		return;
	}

	bssid = (uint8_t *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	if (json_object_object_get_ex(jobj, "bssid", &tmp)) {
		addr = strdup(json_object_get_string(tmp));
		hwaddr_aton(addr, bssid);
		free(addr);
	}

	json_object_put(jobj);
}

static int deagent_get_bssid(struct deagent_private *p, uint8_t *mac,
				const char *ifname, uint8_t *bss_mac)
{
	uint32_t id;
	int res;
	struct blob_buf b = {0};
	char wifi_ap_obj[32] = {0};
	struct sta_query {
		uint8_t sta_mac[6];
		bool is_present;
	} tmp = { 0 };

	snprintf(wifi_ap_obj, 32, "wifi.ap.%s", ifname);

	res = ubus_lookup_id(p->ctx, wifi_ap_obj, &id);
	if (res != UBUS_STATUS_OK) {
		fprintf(stderr, "object '%s' not present", wifi_ap_obj);
		return OBJECT_INVALID;
	}

	blob_buf_init(&b, 0);
	memcpy(tmp.sta_mac, mac, 6);
	tmp.is_present = false;
	res = ubus_invoke(p->ctx, id, "assoclist", b.head,
			check_ap_stalist_cb, (void *)&tmp,
			UBUS_TIMEOUT);
	if (res) {
		fprintf(stderr, "|%s:%d| ubus call |%s assoclist| failed\n",
				__func__, __LINE__, wifi_ap_obj);
		blob_buf_free(&b);
		return -1;
	}

	if (!tmp.is_present)
		goto out;

	blob_buf_free(&b);
	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	res = ubus_invoke(p->ctx, id, "status", b.head,
			get_associated_bss_cb, bss_mac,
			UBUS_TIMEOUT);

	if (res) {
		fprintf(stderr, "|%s:%d| ubus call |%s status| failed\n",
				__func__, __LINE__, wifi_ap_obj);
		blob_buf_free(&b);
		return -1;
	}

out:
	blob_buf_free(&b);
	return 0;

}

/* Handle associated/disassociated STA events */
int deagent_handle_sta_event(struct deagent_private *p, const char *macaddr,
			const char *ifname, const char *event)
{
	int ret = -1;
	uint8_t sta_mac[6] = {0};
	struct wifi_data_element *dm = NULL;
	int status_code = 0;
	int reason_code = 0; // FIXME
	timestamp_t ev_tsp;
	uint8_t bssid[6] = {0};
	uint8_t bss_mac[6] = {0};

	dm = p->dm;
	if (!dm)
		return -1;

	fprintf(stdout, "%s: STA '%s' is %s.\n",
			__func__, macaddr, event);

	/* refresh the data model */
	deagent_getter_run(p);

	hwaddr_aton(macaddr, sta_mac);
	get_timestamp(NULL, ev_tsp);

	if (!strcmp(event, "connected")) {
		ret = sta_eventlog_enqueue(&p->assoc_stalog, ifname, sta_mac);
		if (ret)
			return -1;

		deagent_update_stalog_data(&p->assoc_stalog, ifname, sta_mac);
		deagent_get_bssid(p, sta_mac, ifname, bss_mac);
		deagent_get_assoc_sta(p, ifname, sta_mac, status_code,
				ev_tsp, bss_mac);
		ret |= deagent_fill_sta_bssid(p, sta_mac, bssid);

	} else if (!strcmp(event, "disconnected")) {
		deagent_get_unassoc_sta(p, sta_mac, reason_code, ev_tsp);
		ret = deagent_fill_sta_bssid(p, sta_mac, bssid);
		ret |= sta_eventlog_dequeue(&p->assoc_stalog, sta_mac);
	}

	/* Don't send the notification for single-AP mode, else
	 * Multicast the 1905 topology notification;
	 * send the disassocation stats message in case of
	 * disassociation event (for profile 2 only).
	 */
	if (!ret && !p->cfg.is_sap_mode) {
		prepare_topology_notication(p, sta_mac, event, bssid);
		if ((!strcmp(event, "disconnected")) &&
				(p->cfg.profile == MAP_PROFILE_2)) {
			prepare_disassoc_stats(p, sta_mac, reason_code);
		}
	}

	return 0;

}

int deagent_fill_assoc_frame(struct deagent_private *p, uint8_t *sta_mac,
		uint16_t frame_len, uint8_t *frame)
{
	struct wifi_assoc_frame *new = NULL, *old = NULL;

	new = calloc(1, sizeof(*new));
	if (!new)
		goto error;

	new->len = frame_len;
	new->frame = frame;
	memcpy(new->macaddr, sta_mac, 6);

	/* remove the old frame from framelist
	 * for the client
	 */
	old = wifi_get_frame(p, sta_mac);
	if (old) {
		list_del(&old->list);
		free(old->frame);
		free(old);
	}

	/* add in the framelist */
	list_add(&new->list, &p->framelist);

	return 0;
error:
	free(frame);

	return -1;
}
