/*
 * deagent_ubus.c
 * WiFi Data Elements ubus methods
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <pthread.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#include <easy/easy.h>
#include <wifi.h>

#include "timer.h"
#include <cmdu.h>
#include <1905_tlvs.h>
#include <map2.h>
#include <map_module.h>
#include <cmdu_ackq.h>

#include "eventlog.h"
#include "wfa_delm.h"
#include "deagent.h"
#include "debug.h"


extern int deagent_handle_map_event(struct deagent_private *p,
				struct cmdu_buff *cmdu);
extern int deagent_handle_sta_event(struct deagent_private *p, const char *macaddr,
				const char *ifname, const char *action);
extern int deagent_fill_assoc_frame(struct deagent_private *p, uint8_t *sta_mac,
		uint16_t frame_len, uint8_t *frame);


int deagent_dump(struct ubus_context *ctx, struct ubus_object *obj,
			struct ubus_request_data *req, const char *method,
			struct blob_attr *msg)
{
	struct blob_buf bb;
	struct deagent_private *p = container_of(obj, struct deagent_private, obj);
	int ret;

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);

	ret = deagent_dump_all(p, &bb);
	if (ret) {
		err("DM is not ready; do a refresh!\n");
		goto out;
	}

	ubus_send_reply(ctx, req, bb.head);
out:
	blob_buf_free(&bb);

	return 0;
}

int deagent_refresh(struct ubus_context *ctx, struct ubus_object *obj,
			struct ubus_request_data *req, const char *method,
			struct blob_attr *msg)
{
	struct deagent_private *p = container_of(obj, struct deagent_private, obj);

	if (p->getter_running) {
		fprintf(stderr, "Getter/Collector running!!!\n");
		return -1;
	}

	deagent_getter_run(p);

	return 0;
}

int deagent_event(struct ubus_context *ctx, struct ubus_object *obj,
			struct ubus_request_data *req, const char *method,
			struct blob_attr *msg)
{
	struct blob_buf bb;
	struct deagent_private *p = container_of(obj, struct deagent_private, obj);

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);
	deagent_event_dump(p, &bb);
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

int deagent_publish_object(struct deagent_private *p, const char *objname)
{
	struct ubus_object *obj;
	struct ubus_object_type *obj_type;
	struct ubus_method *obj_methods;
	int ret;
	struct ubus_method m[3] = {
		UBUS_METHOD_NOARG("dump", deagent_dump),
		UBUS_METHOD_NOARG("refresh", deagent_refresh),
		UBUS_METHOD_NOARG("event", deagent_event)
	};
	int num_methods = ARRAY_SIZE(m);

	if (!p->ctx) {
		p->ctx = ubus_connect(NULL);
		if (!p->ctx) {
			fprintf(stderr, "Failed to connect to ubus\n");
			return -1;
		}
		ubus_add_uloop(p->ctx);
	}

	obj = &p->obj;
	memset(obj, 0, sizeof(*obj));

	obj_type = calloc(1, sizeof(struct ubus_object_type));
	if (!obj_type)
		return -1;

	obj_methods = calloc(num_methods, sizeof(struct ubus_method));
	if (!obj_methods) {
		free(obj_type);
		return -1;
	}

	obj->name = objname;
	memcpy(obj_methods, m, num_methods * sizeof(struct ubus_method));
	obj->methods = obj_methods;
	obj->n_methods = num_methods;

	obj_type->name = obj->name;
	obj_type->n_methods = obj->n_methods;
	obj_type->methods = obj->methods;
	obj->type = obj_type;

	ret = ubus_add_object(p->ctx, obj);
	if (ret) {
		fprintf(stderr, "Failed to add 'obj' err = %s\n",
			ubus_strerror(ret));
		free(obj_methods);
		free(obj_type);

		return ret;
	}

	fprintf(stderr, "Added object '%s' id %08x type id %08x\n", objname, obj->id, obj->type->id);

	return 0;
}

void deagent_ieee1905_cmdu_handler(struct deagent_private *priv,
		struct blob_attr *msg)
{
	uint16_t cmdu_type;
	uint16_t mid;
	char in_ifname[16] = {0};
	char src[18] = {0};
	uint8_t src_mac[6] = {0};
	uint8_t origin[6] = {0};
	int len = 0;
	char *tlvstr;
	uint8_t *tlv = NULL;
	struct cmdu_buff *cmdu;
	struct blob_attr *tb[6];
	static const struct blobmsg_policy cmdu_attr[6] = {
		[0] = { .name = "type", .type = BLOBMSG_TYPE_INT16 },
		[1] = { .name = "mid", .type = BLOBMSG_TYPE_INT16 },
		[2] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[3] = { .name = "src", .type = BLOBMSG_TYPE_STRING },
		[4] = { .name = "origin", .type = BLOBMSG_TYPE_STRING },
		[5] = { .name = "cmdu", .type = BLOBMSG_TYPE_STRING },
	};

	blobmsg_parse(cmdu_attr, 6, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1])
		return;

	if (tb[0]) {
		cmdu_type = (uint16_t)blobmsg_get_u16(tb[0]);
		if (cmdu_type > MAP_CMDU_TYPE_MAX)
			return;
	}

	if (tb[1])
		mid = (uint16_t)blobmsg_get_u16(tb[1]);

	if (tb[2])
		strncpy(in_ifname, blobmsg_data(tb[2]), 15);

	if (tb[3]) {
		strncpy(src, blobmsg_data(tb[3]), 17);
		hwaddr_aton(src, src_mac);
	}

	if (tb[4]) {
		memset(src, 0, sizeof(src));
		strncpy(src, blobmsg_data(tb[4]), 17);
		hwaddr_aton(src, origin);
	}

	if (tb[5]) {
		len = blobmsg_data_len(tb[5]) - 16;
		tlvstr = calloc(len + 1, sizeof(char));
		if (!tlvstr)
			return;

		strncpy(tlvstr, blobmsg_data(tb[5]) + 16, len);
		len = (len - 1) / 2;
		tlv = calloc(len, sizeof(uint8_t));
		if (!tlv) {
			free(tlvstr);
			return;
		}

		strtob(tlvstr, len, tlv);
		free(tlvstr);
	}

	cmdu = cmdu_alloc_custom(cmdu_type, &mid, in_ifname,
			src_mac, tlv, len);
	if (!cmdu) {
		fprintf(stderr, "%s: cmdu_alloc_custom failed!\n", __func__);
	} else {
		fprintf(stdout, "%s: cmdu_type:%02x, mid:%u, source:%s\n",
			__func__, cmdu_get_type(cmdu), cmdu_get_mid(cmdu), src);
		memcpy(cmdu->origin, origin, 6);
		deagent_handle_map_event(priv, cmdu);
		cmdu_free(cmdu);
	}

	if (tlv)
		free(tlv);
}

void deagent_sta_event_handler(struct ubus_context *ctx,
				struct ubus_event_handler *ev,
				const char *type,
				struct blob_attr *msg)
{
	char ifname[16] = {0};
	char event[16] = {0};
	struct blob_attr *tb[3];
	static const struct blobmsg_policy ev_attr[3] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	bool add, del;
	struct deagent_private *priv =
		container_of(ev, struct deagent_private, sta_ev);

	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1] || !tb[2])
		return;

	strncpy(ifname,	blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(event, blobmsg_data(tb[1]), sizeof(event) - 1);

	add = !strcmp(event, "connected");
	del = !strcmp(event, "disconnected");

	if (add || del) {
		struct blob_attr *data[2];
		static const struct blobmsg_policy data_attr[2] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "bssid", .type = BLOBMSG_TYPE_STRING },
		};
		char macaddr[18] = {0};
		char bssid[18] = {0};

		blobmsg_parse(data_attr, 2, data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0])
			return;

		strncpy(macaddr, blobmsg_data(data[0]), sizeof(macaddr) - 1);

		if (data[1])
			strncpy(bssid, blobmsg_data(data[1]), sizeof(bssid) - 1);

		deagent_handle_sta_event(priv, macaddr, ifname, event);
	}

	if ((!strcmp(event, "assoc")) || (!strcmp(event, "reassoc"))) {
		char macaddr[18] = {0};
		char *frame_str = NULL;
		uint8_t *frame = NULL;
		uint8_t sta[6] = {0};
		uint16_t len = 0;
		struct blob_attr *data[2];
		static const struct blobmsg_policy data_attr[2] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "raw", .type = BLOBMSG_TYPE_STRING },
		};

		blobmsg_parse(data_attr, 2, data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0] || !data[1])
			return;

		strncpy(macaddr, blobmsg_data(data[0]), sizeof(macaddr) - 1);
		hwaddr_aton(macaddr, sta);

		len = blobmsg_data_len(data[1]);
		frame_str = (char *)calloc(len + 1, sizeof(char));
		if (!frame_str)
			return;

		strncpy(frame_str, blobmsg_data(data[1]), len);
		len = (len - 1)/2;
		frame = (uint8_t *)calloc(len, sizeof(uint8_t));
		if (!frame) {
			free(frame_str);
			return;
		}

		strtob(frame_str, len, frame);
		deagent_fill_assoc_frame(priv, sta, len, frame);
		free(frame_str);
	}
}

void deagent_ubus_object_event_handler(struct ubus_context *ctx,
				struct ubus_event_handler *ev,
				const char *type,
				struct blob_attr *msg)
{
	char path[32] = { 0 };
	struct blob_attr *tb[2];
	static const struct blobmsg_policy ev_attr[2] = {
		[0] = { .name = "id", .type = BLOBMSG_TYPE_INT32 },
		[1] = { .name = "path", .type = BLOBMSG_TYPE_STRING }
	};
	struct deagent_private *p =
		container_of(ev, struct deagent_private, ubus_obj_ev);

	blobmsg_parse(ev_attr, 2, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1])
		return;

	strncpy(path, blobmsg_data(tb[1]), sizeof(path) - 1);

	if (strncmp(path, MAP_PLUGIN_OBJECT,
				strlen(MAP_PLUGIN_OBJECT)))
		return;

	/* update the i1905 & i1905.map object */
	p->i1905 = deagent_lookup_object(p, IEEE1905_OBJECT);
	if (p->i1905 == OBJECT_INVALID) {
		fprintf(stderr, "%s: Failed to get '%s' object!\n",
				__func__, IEEE1905_OBJECT);
		return;
	}

	p->i1905_map = deagent_lookup_object(p, MAP_PLUGIN_OBJECT);
	if (p->i1905_map == OBJECT_INVALID) {
		fprintf(stderr, "%s: Failed to get '%s' object!\n",
				__func__, MAP_PLUGIN_OBJECT);
		return;
	}

	fprintf(stderr, "re-subscribing CMDUs..!!\n");
	deagent_subscribe_for_cmdus(p);
}


void deagent_radio_event_handler(struct ubus_context *ctx,
				struct ubus_event_handler *ev,
				const char *type, struct blob_attr *msg)
{
	int i, idx = 0;
	int noise_dbm;
	uint8_t ch, noise, utilization;
	bool is_intf_found = false;
	bool channel_found = false;
	char ifname[16] = {0};
	char event[16] = {0};
	uint8_t macaddr[6] = {0};
	struct deagent_private *p =
		container_of(ev, struct deagent_private, wifi_radio_ev);
	struct wifi_radio_events *rv = &p->rv;
	struct radio_event_data *r = NULL;
	struct wifi_channel_data *c = NULL;
	struct blob_attr *tb[3];
	struct blob_attr *data[3];
	static const struct blobmsg_policy ev_attr[3] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	static const struct blobmsg_policy data_attr[3] = {
		[0] = { .name = "channel", .type = BLOBMSG_TYPE_INT32 },
		[1] = { .name = "noise", .type = BLOBMSG_TYPE_INT32 },
		[2] = { .name = "utilization", .type = BLOBMSG_TYPE_INT32 },
	};

	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1] || !tb[2])
		return;

	strncpy(ifname, blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(event, blobmsg_data(tb[1]), sizeof(event) - 1);
	if (wifi_radio_get_hwaddr(ifname, macaddr) != 0)
		if_gethwaddr(ifname, macaddr);

	/* check the event type */
	if (strncmp(event, "survey", sizeof(event) - 1))
		return;

	if (hwaddr_is_zero(macaddr))
		return;

	for (i = 0; i < rv->num_radios; i++) {
		r = rv->rdata + i;

		if (hwaddr_equal(r->macaddr, macaddr)) {
			is_intf_found = true;
			break;
		}
	}

	/* Add the entry for new interface */
	if (!is_intf_found) {
		rv->rdata = (struct radio_event_data *)realloc(rv->rdata,
				(rv->num_radios + 1)*sizeof(*r));

		if (!rv->rdata) {
			fprintf(stderr, "%s: %d -ENOMEM\n",
					__func__, __LINE__);
			return;
		}

		idx = rv->num_radios;
		memset(&rv->rdata[idx], 0, sizeof(*r));
		memcpy(rv->rdata[idx].macaddr, macaddr, 6);
		r = rv->rdata + idx;
		rv->num_radios++;
	}

	blobmsg_parse(data_attr, 3, data, blobmsg_data(tb[2]),
			blobmsg_data_len(tb[2]));

	ch = (uint8_t)blobmsg_get_u32(data[0]);
	noise_dbm = blobmsg_get_u32(data[1]);
	noise = rssi_to_rcpi(noise_dbm);
	utilization = (uint8_t)blobmsg_get_u32(data[2]);


	for (i = 0; i < r->num_cdata; i++) {
		c = r->cdata + i;

		if (c->channel == ch) {
			channel_found = true;
			c->anpi = noise;
			c->utilization = utilization;
			break;
		}
	}

	/* Add the entry if not found in cdata */
	if (!channel_found) {
		r->cdata = (struct wifi_channel_data *)realloc(r->cdata,
				(r->num_cdata + 1)*sizeof(*c));
		if (!r->cdata) {
			fprintf(stderr, "%s: %d -ENOMEM\n",
					__func__, __LINE__);
			return;
		}

		idx = r->num_cdata;
		memset(&r->cdata[idx], 0, sizeof(struct wifi_channel_data));
		r->cdata[idx].channel = ch;
		r->cdata[idx].anpi = noise;
		r->cdata[idx].utilization = utilization;
		r->num_cdata++;
	}
}

int deagent_register_events(struct deagent_private *p)
{
	int ret = 0;

	p->sta_ev.cb = deagent_sta_event_handler;
	ret = ubus_register_event_handler(p->ctx, &p->sta_ev, "wifi.sta");
	if (ret) {
		fprintf(stderr, "deagent_register_event() failed %s\n for wifi.sta",
				ubus_strerror(ret));
	}

	p->wifi_radio_ev.cb = deagent_radio_event_handler;
	ret = ubus_register_event_handler(p->ctx, &p->wifi_radio_ev, "wifi.radio");
	if (ret) {
		fprintf(stderr, "deagent_register_event() failed %s\n for wifi.radio",
				ubus_strerror(ret));
	}

	p->ubus_obj_ev.cb = deagent_ubus_object_event_handler;
	ret = ubus_register_event_handler(p->ctx, &p->ubus_obj_ev, "ubus.object.add");
	if (ret) {
		fprintf(stderr, "deagent_register_event() failed %s\n for ubus_object_add",
				ubus_strerror(ret));
	}

	return ret;
}

uint32_t deagent_lookup_object(struct deagent_private *p, const char *objname)
{
	uint32_t id;
	int status;

	status = ubus_lookup_id(p->ctx, objname, &id);
	if (status != UBUS_STATUS_OK) {
		fprintf(stderr, "object '%s' not present!\n", objname);
		return OBJECT_INVALID;
	}

	return id;
}

static void ieee1905_info_response_cb(struct ubus_request *req,
					int mtype, struct blob_attr *msg)
{
	struct json_object *jobj = NULL;
	struct json_object *mac;
	struct deagent_private *p;
	char *addr = NULL;
	char *str;

	if (!msg || !req->priv) {
		fprintf(stderr, "%s:Message recieved is NULL\n", __func__);
		return;
	}

	p = (struct deagent_private *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	/* Store self device mac addr */
	if (json_object_object_get_ex(jobj, "ieee1905id", &mac)) {
		addr = strdup(json_object_get_string(mac));
		hwaddr_aton(addr, p->ieee1905_macaddr);
		free(addr);
	}

	json_object_put(jobj);
}

int deagent_get_ieee1905id(struct deagent_private *p)
{
	struct blob_buf bb = { 0 };
	int ret = 0;

	blob_buf_init(&bb, 0);
	ret = ubus_invoke(p->ctx, p->i1905, "info", bb.head,
				ieee1905_info_response_cb, p, 1000);
	if (ret)
		fprintf(stderr, "Failed to get i1905 info (err = %s)\n",
						ubus_strerror(ret));

	blob_buf_free(&bb);

	return ret;
}

static void self_node_topology_resp_cb(struct ubus_request *req,
		int type, struct blob_attr *msg)
{
	struct json_object *jobj = NULL;
	struct json_object *tmp;
	struct deagent_private *priv;
	char *ieee1905_data = NULL;
	uint8_t *tlv = NULL;
	int len;
	char *str;
	struct wifi_network_device *own = NULL;

	if (!msg || !req->priv) {
		fprintf(stderr, "%s:Message recieved is NULL\n", __func__);
		return;
	}

	priv = (struct deagent_private *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	if (json_object_object_get_ex(jobj, "data", &tmp)) {
		ieee1905_data = strdup(json_object_get_string(tmp));
		fprintf(stderr, "i1905 topo resp tlv data: %s\n", ieee1905_data);
		len = strlen(ieee1905_data);
		len = (len - 1)/2;
		tlv = calloc(len, sizeof(uint8_t));
		if (!tlv) {
			fprintf(stderr, "%s: -ENOMEM\n", __func__);
			free(ieee1905_data);
			goto error;
		}

		strtob(ieee1905_data, len, tlv);
		free(ieee1905_data);

		if ((!priv->getter_running) && priv->dm) {
			own = &priv->dm->network.device[0];
			if (own->ieee1905_tlv_data) {
				free(own->ieee1905_tlv_data);
				own->ieee1905_tlv_data = NULL;
			}

			own->ieee1905_tlv_data = tlv;
			own->ieee1905_tlv_datalen = len;
		}
	}

error:
	json_object_put(jobj);
}

int self_node_topology_query(struct deagent_private *priv)
{
	struct blob_buf bb;
	int ret = 1;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);
	blobmsg_add_u32(&bb, "type", CMDU_TYPE_TOPOLOGY_RESPONSE);
	ret = ubus_invoke(priv->ctx, priv->i1905, "buildcmdu", bb.head,
			self_node_topology_resp_cb, priv, 1000);
	if (ret)
		fprintf(stderr, "Failed to get 'ieee1905' query_topology\n");

	blob_buf_free(&bb);

	return ret;
}

int deagent_map_sub_cb(void *bus, void *priv, void *data)
{
	struct blob_attr *msg = (struct blob_attr *)data;
	char *str;


	str = blobmsg_format_json(msg, true);
	fprintf(stdout, "deagent: Received: %s\n", str);
	free(str);

	deagent_ieee1905_cmdu_handler(priv, msg);
	return 0;
}

int deagent_map_del_cb(void *bus, void *priv, void *data)
{
	uint32_t *obj = (uint32_t *)data;

	fprintf(stdout, "Object 0x%x no longer present\n", *obj);
	return 0;
}

int deagent_subscribe_for_cmdus(struct deagent_private *priv)
{
	mapmodule_cmdu_mask_t cmdu_mask = {0};
	int ret;


	map_prepare_cmdu_mask(cmdu_mask,
			CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE,
			CMDU_TYPE_TOPOLOGY_QUERY,
			CMDU_AP_CAPABILITY_QUERY,
			CMDU_AP_METRICS_QUERY,
			CMDU_CLIENT_CAPABILITY_QUERY,
			CMDU_UNASSOC_STA_LINK_METRIC_QUERY,
			CMDU_BACKHAUL_STA_CAPABILITY_QUERY,
			CMDU_CHANNEL_SCAN_REQUEST,
			CMDU_CHANNEL_SELECTION_REQ,
			CMDU_TYPE_INVALID);


	ret = map_subscribe(priv->ctx,
			    &priv->i1905_map,
			    "deagent", &cmdu_mask, priv,
			    deagent_map_sub_cb,
			    deagent_map_del_cb,
			    &priv->subscriber);

	return ret;
}
