/*
 * decollector_cmdu.c
 * File contains CMDUs handling needed for WiFi Data Elements.
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: saurabh.verma@iopsys.eu
 *	   anjan.chanda@iopsys.eu
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <net/if_arp.h>
#include <pthread.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <net/if.h>

#include <easy/easy.h>

#include "timer.h"
#include <cmdu.h>
#include <1905_tlvs.h>
#include <map2.h>
#include <cmdu_ackq.h>

#include "wfa_delm.h"
#include "decollector.h"
#include "tlv.h"
#include "common.h"
#include "utils.h"
#include "cmdu_validate.h"

#define UBUS_TIMEOUT		1000
#define FRAME_SIZE		4800
#define min(x, y)       	(x) < (y) ? (x) : (y)
#define IE_HT_CAP		45
#define IE_VHT_CAP		191
#define IE_EXT_HE_CAP		35



static void decollector_send_cmdu_cb(struct ubus_request *req,
			      int type, struct blob_attr *msg)
{
	struct json_object *jobj = NULL;
	struct json_object *tmp;
	uint16_t *mid;
	char *str;

	if (!msg || !req->priv) {
		fprintf(stderr, "%s:Message recieved is NULL\n", __func__);
		return;
	}

	mid = (uint16_t *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	if (json_object_object_get_ex(jobj, "mid", &tmp))
		*mid = json_object_get_int(tmp);


	json_object_put(jobj);
}

int prepare_send_cmdu_blob(struct cmdu_buff *cmdu_data,
		struct blob_buf *b)
{
	char *blob_data = NULL;
	char dst_addr[18] = { 0 };

	blobmsg_add_u32(b, "type", cmdu_get_type(cmdu_data));
	blobmsg_add_u32(b, "mid", cmdu_get_mid(cmdu_data));

	hwaddr_ntoa(cmdu_data->origin, dst_addr);
	blobmsg_add_string(b, "dst", dst_addr);

	fprintf(stdout, "|%s:%d|cmdu:%02x|dst:%s|\n", __func__, __LINE__,
			cmdu_get_type(cmdu_data), dst_addr);

	if (cmdu_data->datalen) {
		char *tlv_str = NULL;
		uint16_t len = 0;

		len = (cmdu_data->datalen * 2) + 1;

		tlv_str = calloc(len, sizeof(char));
		if (!tlv_str)
			goto out;

		btostr(cmdu_data->data, cmdu_data->datalen, tlv_str);
		tlv_str[len-1] = '\0';

		blobmsg_add_string(b, "data", tlv_str);
		free(tlv_str);
	}

	blob_data = blobmsg_format_json(b->head, true);
	if (blob_data) {
		fprintf(stderr, "blob buf send data: |%s|\n", blob_data);
		free(blob_data);
	}

	return 0;
out:

	return -1;
}

int decollector_send_cmdu_request(struct decollector_private *priv, struct cmdu_buff *req,
			   uint16_t resp_type)
{
	int ret = 0;
	uint16_t msgid = 0;
	struct blob_buf b = { 0 };
	uint16_t message_type = cmdu_get_type(req);

	fprintf(stderr, "%s: send CMDU 0x%4x\n", __func__, message_type);

	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);
	ret = prepare_send_cmdu_blob(req, &b);
	if (ret)
		goto out;

	ret = ubus_invoke(priv->ctx, priv->i1905, "cmdu",
			  b.head, decollector_send_cmdu_cb,
			  (void *)&msgid,
			  UBUS_TIMEOUT);
	if (!ret) {
		fprintf(stderr, "%s: msg-id = %hu\n", __func__, msgid);

		/* Don't enqueue message id for autoconfig search &
		 * channel selection request & client capability query
		 */
		if ((message_type == CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH) ||
				(message_type == CMDU_CHANNEL_SELECTION_REQ) ||
				(message_type == CMDU_CLIENT_CAPABILITY_QUERY))
			goto out;

		/* Store mid of the sent CMDU with an expected response type.
		 * On receiving a 'resp_type' cmdu, lookup in the ackq and
		 * drop or handle it.
		 */
		cmdu_ackq_enqueue(&priv->cmdu_ack_q, resp_type, msgid,
				  req->origin, 2000, 0, NULL);  //FIXME: tmo
	} else {
		fprintf(stderr, "[%s:%d] ubus call failed for |ieee1905 cmdu|",
				__func__, __LINE__);

		/* unset getter_running,
		 * if not set, refresh method won't work
		 */
		priv->getter_running = 0;

		/* update the i1905 object */
		priv->i1905 = decollector_lookup_object(priv, IEEE1905_OBJECT);
		if (priv->i1905 == OBJECT_INVALID)
			fprintf(stderr, "Failed to get '%s' object!\n",
					IEEE1905_OBJECT);
	}

out:
	blob_buf_free(&b);

	return ret;
}

static int get_sta_index(struct wifi_bss_element *bss, uint8_t *sta)
{
	int i = 0;
	char addr[18] = {0};

	for (i = 0; i < bss->num_stations; i++) {
		if (hwaddr_equal(bss->stalist[i].macaddr, sta)) {
			hwaddr_ntoa(sta, addr);
			fprintf(stdout, "found sta:|%s| at index %d\n", addr, i);
			return i;
		}
	}
	return -1;
}

static int get_opclasss_index(struct wifi_radio_element *radio, uint8_t opclass_id)
{
	int i = 0;
	struct wifi_scanres_opclass_element *opclass = NULL;

	/* TODO: Assuming single scanresult per radio */
	if (radio->scanlist) {
		for (i = 0; i < radio->scanlist[0].num_opclass_scanned; i++) {
			opclass = radio->scanlist[0].opclass_scanlist + i;
			if (opclass->opclass == opclass_id)
				return i;
		}
	}

	return -1;
}

static int get_channel_index(struct wifi_scanres_opclass_element *opclass, uint8_t channel_id)
{
	int i = 0;
	struct wifi_scanres_channel_element *ch = NULL;

	for (i = 0; i < opclass->num_channels_scanned; i++) {
		ch = opclass->channel_scanlist + i;
		if (ch->channel == channel_id)
			return i;
	}

	return -1;
}

static int get_radio_index(struct wifi_network_device *dev, uint8_t *radio_id)
{
	int i = 0;
	char addr[18] = {0};

	for (i = 0; i < dev->num_radios; i++) {
		if (hwaddr_equal(dev->radiolist[i].macaddr, radio_id)) {
			hwaddr_ntoa(radio_id, addr);
			fprintf(stdout, "found radio:|%s| at index %d\n", addr, i);
			return i;
		}
	}
	return -1;
}

static int get_bss_index(struct wifi_radio_element *radio, uint8_t *bssid)
{
	int i = 0;
	char addr[18] = {0};

	for (i = 0; i < radio->num_bss; i++) {
		if (hwaddr_equal(radio->bsslist[i].bssid, bssid)) {
			hwaddr_ntoa(bssid, addr);
			fprintf(stdout, "found bss:|%s| at index %d\n", addr, i);
			return i;
		}
	}
	return -1;
}

static int get_radio_bss_sta_index(struct wifi_network_device *dev, uint8_t *sta,
		int *radio_index, int *bss_index)
{
	int i, j;
	int sta_index;

	for (i = 0; i < dev->num_radios; i++) {
		for (j = 0; j < dev->radiolist[i].num_bss; j++) {
			sta_index = get_sta_index(&dev->radiolist[i].bsslist[j], sta);
			if (sta_index != -1) {
				*radio_index = i;
				*bss_index = j;
				return sta_index;
			}
		}
	}

	return -1;
}

static int get_radio_and_bss_index(struct wifi_network_device *dev,
		uint8_t *bssid, int *radio_index)
{
	int i;
	int bss_index;

	for (i = 0; i < dev->num_radios; i++) {
		bss_index = get_bss_index(&dev->radiolist[i], bssid);
		if (bss_index != -1) {
			*radio_index = i;
			return bss_index;
		}
	}

	return -1;
}

static struct wifi_radio_element* get_radio_from_opclass(
		struct wifi_network_device *dev, uint8_t opclass_id)
{
	int i, j;
	struct wifi_radio_element *radio = NULL;
	struct wifi_opclass_current_element *curr_opclass = NULL;

	for (i = 0; i < dev->num_radios; i++) {
		radio = dev->radiolist + i;

		for (j = 0; j < radio->num_curr_opclass; j++) {
			curr_opclass = radio->curr_opclass + j;

			if (curr_opclass->id == opclass_id)
				return radio;
		}
	}

	return NULL;
}

struct wifi_network_device *decollector_get_origin_dev(struct decollector_private *p,
		struct wifi_data_element *dm, uint8_t *origin)
{
	int i;

	if (!dm) {
		fprintf(stderr, "%s: p->dm is NULL\n", __func__);
		return NULL;
	}

	fprintf(stderr, "%s: num_devices = %u\n", __func__, dm->network.num_devices);
	for (i = 0; i < dm->network.num_devices; i++) {
		if (hwaddr_equal(dm->network.device[i].macaddr, origin))
			return &dm->network.device[i];
	}

	return NULL;
}

static bool fill_topology_resp_from_tlv(struct decollector_private *priv,
		struct cmdu_buff *resp, struct wifi_network_device *dev)
{
	fprintf(stdout, "%s: %d\n", __func__, __LINE__);

	int i, j;
	int offset = 0;
	int radio_index = -1;
	int bss_index = -1;
	struct wifi_radio_element *radio;
	struct wifi_bss_element *bss;
	struct wifi_sta_element *sta;
	struct tlv *tv[3][16];

	/* Set device default profile to profile-1 */
	dev->map_profile = MAP_PROFILE_1;

	if (!validate_topology_response(resp, tv)) {
		fprintf(stderr, "%s: cmdu validation:" \
				"[TOPOLOGY_RESPONSE] failed\n", __func__);
		return false;
	}

	/* Parse AP Operational BSS TLV */
	if (tv[0][0]) {
		uint8_t *tv_data = (uint8_t *)tv[0][0]->data;
		struct tlv_ap_oper_bss *p =
			(struct tlv_ap_oper_bss *)tv_data;

		dev->num_radios = p->num_radio;
		offset = sizeof(*p);
		if (dev->num_radios > 0)
			dev->radiolist = calloc(dev->num_radios,
					sizeof(struct wifi_radio_element));
		if (!dev->radiolist) {
			fprintf(stderr, "%s:%d -ENOMEM\n", __func__, __LINE__);
			dev->num_radios = 0;
			return false;
		}

		for (i = 0; i < dev->num_radios; i++) {
			struct ap_oper_bss_radiolist *r =
				(struct ap_oper_bss_radiolist *)&tv_data[offset];

			radio = dev->radiolist + i;
			memcpy(radio->macaddr, r->radio, 6);
			radio->enabled = true;
			radio->num_bss = r->num_bss;
			offset += sizeof(*r);

			if (radio->num_bss > 0) {
				radio->bsslist = calloc(radio->num_bss,
						sizeof(struct wifi_bss_element));

				if (!radio->bsslist) {
					fprintf(stderr, "%s:%d -ENOMEM\n",
							__func__, __LINE__);
					radio->num_bss = 0;
					return false;
				}
			}

			for (j = 0; j < radio->num_bss; j++) {
				struct ap_oper_bss_bss *b =
					(struct ap_oper_bss_bss *)&tv_data[offset];

				bss = radio->bsslist + j;
				memcpy(bss->bssid, b->bssid, 6);
				memcpy(bss->ssid, b->ssid, b->ssidlen);
				fprintf(stderr, "%s: %d === ssid = %s\n", __func__, j, bss->ssid);
				bss->enabled = true;
				get_timestamp(NULL, bss->tsp);
				offset += sizeof(*b) + b->ssidlen;
			}
		}
	}

	/* Parse Associated Clients TLV */
	if (tv[1][0]) {
		uint8_t *tv_data = (uint8_t *)tv[1][0]->data;
		struct tlv_assoc_client *p =
			(struct tlv_assoc_client *)tv_data;

		offset = sizeof(*p);
		for (i = 0; i < p->num_bss; i++) {
			struct assoc_client_bss *b =
				(struct assoc_client_bss *)&tv_data[offset];

			offset += sizeof(*b);
			bss_index = get_radio_and_bss_index(dev,
					b->bssid, &radio_index);
			if (bss_index == -1)
				continue;

			radio = dev->radiolist + radio_index;
			bss = radio->bsslist + bss_index;
			bss->num_stations = BUF_GET_BE16(b->num_client);
			if (bss->num_stations > 0) {
				bss->stalist = calloc(bss->num_stations,
						sizeof(struct wifi_sta_element));

				if (!bss->stalist) {
					fprintf(stderr, "%s:%d -ENOMEM\n",
							__func__, __LINE__);
					bss->num_stations = 0;
					return false;
				}
			}


			for (j = 0; j < bss->num_stations; j++) {
				struct assoc_client_sta *s =
					(struct assoc_client_sta *)&tv_data[offset];

				sta = bss->stalist + j;
				memcpy(sta->macaddr, s->macaddr, 6);
				sta->conn_time = BUF_GET_BE16(s->conntime);
				get_timestamp(NULL, sta->tsp);
				offset += sizeof(*s);
			}
		}
	}

	/* Parse Multi-AP Profile TLV */
	if (tv[2][0]) {
		struct tlv_map_profile *p =
			(struct tlv_map_profile *)tv[2][0]->data;

		dev->map_profile = p->profile;
	}

	return true;
}

static bool fill_ap_capability_from_tlv(struct decollector_private *priv,
		struct cmdu_buff *resp, struct wifi_network_device *dev)
{
	int i;
	int index = 0;
	int offset = 0;
	int radio_index = -1;
	struct wifi_radio_element *radio;
	struct wifi_opclass_supported_element *opclass;
	struct tlv *tv[9][16];

	if (!validate_ap_cap_report(resp, tv)) {
		fprintf(stderr, "%s: cmdu validation:" \
				"[AP_CAPABILITY_REPORT] failed\n", __func__);
		return false;
	}

	/* AP Capability TLV */
	if (tv[0][0]) {
		struct tlv_ap_cap *p = (struct tlv_ap_cap *)tv[0][0]->data;

		/* TODO: Currently no flag described in the
		 * wifi network device/radio
		 */
		if (!(p->cap & UNASSOC_STA_REPORTING_ONCHAN)) {
			for (i = 0; i < dev->num_radios; i++) {
				dev->radiolist[i].num_unassoc_sta = 0;
				dev->radiolist[i].unassoc_stalist = NULL;
			}
		}
	}

	index = 0;
	/* Parse AP Radio Basic Capabilities TLV */
	while (tv[1][index] && (index < 16)) {
		uint8_t *tv_data = (uint8_t *)tv[1][index++]->data;
		struct tlv_ap_radio_basic_cap *p =
			(struct tlv_ap_radio_basic_cap *)tv_data;

		offset = sizeof(*p);
		radio_index = get_radio_index(dev, p->radio);
		if (radio_index == -1)
			continue;

		radio = dev->radiolist + radio_index;
		radio->num_supp_opclass = p->num_opclass;
		if (radio->num_supp_opclass > 0) {
			radio->supp_opclass = calloc(radio->num_supp_opclass,
					sizeof(struct wifi_opclass_supported_element));

			if (!radio->supp_opclass) {
				fprintf(stderr, "%s:%d -ENOMEM\n",
						__func__, __LINE__);
				radio->num_supp_opclass = 0;
				return false;
			}
		}

		for (i = 0; i < radio->num_supp_opclass; i++) {
			struct ap_radio_basic_cap_opclass *op =
				(struct ap_radio_basic_cap_opclass *)&tv_data[offset];
			struct wifi_opclass_entry e;
			int j;

			opclass = radio->supp_opclass + i;
			opclass->id = op->classid;
			opclass->max_txpower = op->max_txpower;
			opclass->num_exclude_channels = op->num_nonop_channel;


			if (!wifi_get_opclass_entry(opclass->id, &e)) {
				if (e.bw == BW20) {
					opclass->num_channels = e.ch.num;
					opclass->chanlist = calloc(sizeof(uint8_t), opclass->num_channels);
					if (opclass->chanlist) {
						for (j = 0; j < opclass->num_channels; j++)
							opclass->chanlist[j] = e.ch.ch[j];
					}
				}
			}

			if (opclass->num_exclude_channels > 0)
				opclass->exclude_chanlist = calloc(sizeof(uint8_t),
						opclass->num_exclude_channels);

			if (opclass->exclude_chanlist)
				memcpy(opclass->exclude_chanlist,
						op->nonop_channel,
						opclass->num_exclude_channels);

			offset += sizeof(*op) + op->num_nonop_channel;
		}
	}

	index = 0;
	/* Parse AP HT Capabilities TLV */
	while (tv[2][index] && (index < 16)) {
		struct tlv_ap_ht_cap *p =
			(struct tlv_ap_ht_cap *)tv[2][index++]->data;

		radio_index = get_radio_index(dev, p->radio);
		if (radio_index == -1)
			continue;

		radio = dev->radiolist + radio_index;
		memcpy(&radio->caps.ht, &p->cap, 1);
		radio->caps.valid |= HT_CAP_VALID;
	}

	index = 0;
	/* AP VHT Capabilities TLV */
	while (tv[3][index] && (index < 16)) {
		uint8_t vht[6];
		struct tlv_ap_vht_cap *p =
			(struct tlv_ap_vht_cap *)tv[3][index++]->data;

		radio_index = get_radio_index(dev, p->radio);
		if (radio_index == -1)
			continue;

		memset(vht, 0, sizeof(vht));
		BUF_PUT_BE16(vht[0], p->tx_mcs_supported);
		BUF_PUT_BE16(vht[2], p->rx_mcs_supported);
		memcpy(&vht[4], p->cap, 2);

		radio = dev->radiolist + radio_index;
		memcpy(radio->caps.vht, vht, sizeof(vht));
		radio->caps.valid |= VHT_CAP_VALID;
	}

	index = 0;
	/* Parse AP HE Capabilities TLV */
	while (tv[4][index] && (index < 16)) {
		int c_index = 0;
		uint8_t he[15] = {0};
		uint8_t *tv_data = (uint8_t *)tv[4][index++]->data;
		struct tlv_ap_he_cap *p =
			(struct tlv_ap_he_cap *)tv_data;

		offset = 0;
		radio_index = get_radio_index(dev, p->radio);
		if (radio_index == -1)
			continue;

		if (p->hemcs.len > 12)
			continue;

		offset = 1 + 6;
		he[0] = p->hemcs.len;

		if (p->hemcs.len > 0)
			memcpy(&he[1], &tv_data[offset], p->hemcs.len);

		c_index = 1 + he[0];
		if (c_index >= sizeof(he) - 2)
			continue;

		offset += he[0];
		memcpy(&he[c_index], &tv_data[offset], 2);

		radio = dev->radiolist + radio_index;
		memcpy(radio->caps.he, he, sizeof(he));
		radio->caps.valid |= HE_CAP_VALID;
	}

	/* Parse Metric Collection Interval TLV */
	if (tv[5][0]) {
		/* TODO
			struct tlv_metric_collection_int *p =
			(struct tlv_metric_collection_int *)tv[5][0]->data;
		*/
		dev->collect_int = priv->cfg.refresh_int;
	}

	return true;
}

static bool fill_apstats_from_tlv(struct decollector_private *priv,
		struct cmdu_buff *resp, struct wifi_network_device *dev)
{
	int i;
	int index, offset = 0;
	struct wifi_radio_element *radio;
	struct wifi_bss_element *bss;
	struct wifi_sta_element *sta;
	int radio_index;
	int bss_index;
	int sta_index;
	struct tlv *tv[6][16];

	if (!validate_ap_metric_resp(resp, tv)) {
		fprintf(stderr, "%s: cmdu validation" \
				"[AP_METRIC_RESP] failed\n", __func__);
		return false;
	}

	index = 0;
	/* Parse AP Metrics TLV */
	while (tv[0][index] && (index < 16)) {
		int copy_index;
		struct tlv_ap_metrics *p =
			(struct tlv_ap_metrics *)tv[0][index++]->data;

		bss_index = get_radio_and_bss_index(dev, p->bssid, &radio_index);
		if (bss_index == -1)
			continue;

		/* TODO/revisit:
		 * channel utilization & totoal STAs
		 */
		radio = dev->radiolist + radio_index;
		bss = radio->bsslist + bss_index;
		copy_index = 0;
		if (p->esp_ac & ESP_AC_BE)
			memcpy(bss->est_wmm_be, p->esp_be, 3);

		if (p->esp_ac & ESP_AC_BK) {
			memcpy(bss->est_wmm_bk, p->esp + copy_index, 3);
			copy_index += 3;
		}

		if (p->esp_ac & ESP_AC_VO) {
			memcpy(bss->est_wmm_vo, p->esp + copy_index, 3);
			copy_index += 3;
		}

		if (p->esp_ac & ESP_AC_VI)
			memcpy(bss->est_wmm_vi, p->esp + copy_index, 3);
	}

	index = 0;
	/* Parse Associated STA Traffic Stats TLV */
	while (tv[1][index] && (index < 16)) {
		struct tlv_assoc_sta_traffic_stats *p =
			(struct tlv_assoc_sta_traffic_stats *)tv[1][index++]->data;

		sta_index = get_radio_bss_sta_index(dev, p->macaddr,
				&radio_index, &bss_index);
		if (sta_index == -1)
			continue;

		radio = dev->radiolist + radio_index;
		bss = radio->bsslist + bss_index;
		sta = bss->stalist + sta_index;

		sta->tx_bytes = BUF_GET_BE32(p->tx_bytes);
		sta->rx_bytes = BUF_GET_BE32(p->rx_bytes);
		sta->tx_pkts = BUF_GET_BE32(p->tx_packets);
		sta->rx_pkts = BUF_GET_BE32(p->rx_packets);
		sta->tx_errors = BUF_GET_BE32(p->tx_err_packets);
		sta->rx_errors = BUF_GET_BE32(p->rx_err_packets);
		sta->rtx_pkts = BUF_GET_BE32(p->rtx_packets);
	}

	index = 0;
	/* Parse Associated STA Link Metrics TLV */
	while (tv[2][index] && (index < 16)) {
		uint8_t *tv_data = (uint8_t *)tv[2][index++]->data;
		struct tlv_assoc_sta_link_metrics *p =
			(struct tlv_assoc_sta_link_metrics *)tv_data;

		offset = sizeof(*p);
		for (i = 0; i < p->num_bss; i++) {
			struct assoc_sta_link_metrics_bss *b =
				(struct assoc_sta_link_metrics_bss *)&tv_data[offset];

			bss_index = get_radio_and_bss_index(dev,
					b->bssid, &radio_index);
			offset += sizeof(*b);
			if (bss_index == -1)
				continue;

			radio = dev->radiolist + radio_index;
			bss = radio->bsslist + bss_index;
			sta_index = get_sta_index(bss, p->macaddr);
			if (sta_index != -1) {
				sta = bss->stalist + sta_index;
				sta->dl_est_thput = BUF_GET_BE32(b->dl_thput);
				sta->ul_est_thput = BUF_GET_BE32(b->ul_thput);
				sta->rssi = b->ul_rcpi;
			}

		}
	}

	index = 0;
	/* Parse AP Extended Metrics TLV */
	while (tv[3][index] && (index < 16)) {
		struct tlv_ap_ext_metrics *p =
			(struct tlv_ap_ext_metrics *)tv[3][index++]->data;

		bss_index = get_radio_and_bss_index(dev, p->bssid, &radio_index);
		if (bss_index == -1)
			continue;

		radio = dev->radiolist + radio_index;
		bss = radio->bsslist + bss_index;

		bss->tx_ucast_bytes = BUF_GET_BE32(p->tx_bytes_ucast);
		bss->rx_ucast_bytes = BUF_GET_BE32(p->rx_bytes_ucast);
		bss->tx_mcast_bytes = BUF_GET_BE32(p->tx_bytes_mcast);
		bss->rx_mcast_bytes = BUF_GET_BE32(p->rx_bytes_mcast);
		bss->tx_bcast_bytes = BUF_GET_BE32(p->tx_bytes_bcast);
		bss->rx_bcast_bytes = BUF_GET_BE32(p->rx_bytes_bcast);
	}

	index = 0;
	/* Parse Radio Metrics TLV */
	while (tv[4][index] && (index < 16)) {
		struct tlv_radio_metrics *p =
			(struct tlv_radio_metrics *)tv[4][index++]->data;

		radio_index = get_radio_index(dev, p->radio);
		if (radio_index == -1)
			continue;

		radio = dev->radiolist + radio_index;
		radio->anpi = p->noise;
		radio->tx_utilization = p->transmit;
		radio->rx_utilization = p->receive_self;
		radio->other_utilization = p->receive_other;
	}

	index = 0;
	/* Parse Associated STA Extended Link Metrics TLV */
	while (tv[5][index] && (index < 16)) {
		uint8_t *tv_data = (uint8_t *)tv[5][index++]->data;
		struct tlv_sta_ext_link_metric *p =
			(struct tlv_sta_ext_link_metric *)tv_data;

		offset = sizeof(*p);
		for (i = 0; i < p->num_bss; i++) {
			struct sta_ext_link_metric_bss *b =
				(struct sta_ext_link_metric_bss *)&tv_data[offset];

			bss_index = get_radio_and_bss_index(dev,
					b->bssid, &radio_index);
			offset += sizeof(*b);
			if (bss_index == -1)
				continue;

			radio = dev->radiolist + radio_index;
			bss = radio->bsslist + bss_index;
			sta_index = get_sta_index(bss, p->macaddr);
			if (sta_index != -1) {
				sta = bss->stalist + sta_index;
				sta->dl_rate = BUF_GET_BE32(b->dl_rate);
				sta->ul_rate = BUF_GET_BE32(b->ul_rate);
				sta->dl_utilization = BUF_GET_BE32(b->rx_util);
				sta->ul_utilization = BUF_GET_BE32(b->tx_util);
			}
		}
	}

	return true;
}

static bool fill_bksta_from_tlv(struct decollector_private *priv,
		struct cmdu_buff *resp, struct wifi_network_device *dev)
{
	int index;
	int radio_index;
	struct wifi_radio_element *radio;
	struct tlv *tv[1][16];

	if (!validate_bksta_cap_report(resp, tv)) {
		fprintf(stderr, "%s: cmdu validation " \
				"[BACKHAUL_STA_CAPABILITY_REPORT] failed\n",
				__func__);
		return false;
	}

	index = 0;
	/* Parse Backhaul STA Radio Capabilities TLV */
	while (tv[0][index] && (index < 16)) {
		struct tlv_bsta_radio_cap *p =
			(struct tlv_bsta_radio_cap *)tv[0][index++]->data;

		radio_index = get_radio_index(dev, p->radio);
		if (radio_index == -1)
			continue;

		radio = dev->radiolist + radio_index;
		if (p->macaddr_included & BSTA_MACADDRESS_INCLUDED)
			memcpy(radio->bksta.macaddr, p->macaddr[0], 6);
	}

	return true;
}

static bool fill_unassocsta_from_tlv(struct decollector_private *priv,
		struct cmdu_buff *resp, struct wifi_network_device *dev)
{
	int i;
	struct wifi_radio_element *radio = NULL;
	struct wifi_unassoc_sta_element *unassoc_sta;
	struct tlv *tv[1][16];

	if (!validate_unassoc_sta_link_metric_resp(resp, tv)) {
		fprintf(stderr, "%s: cmdu validation " \
				"[UNASSOC_STA_LINK_METRIC_RESP] failed\n",
				__func__);
		return false;
	}

	/* Parse Unassociated STA Link Metrics response TLV */
	if (tv[0][0]) {
		struct tlv_unassoc_sta_link_metrics_resp *p =
			(struct tlv_unassoc_sta_link_metrics_resp *)tv[0][0]->data;


		radio = get_radio_from_opclass(dev, p->opclass);
		if (radio == NULL)
			return false;

		radio->num_unassoc_sta = p->num_sta;

		if (radio->num_unassoc_sta > 0) {
			radio->unassoc_stalist = calloc(radio->num_unassoc_sta,
					sizeof(struct wifi_unassoc_sta_element));

			if (!radio->unassoc_stalist) {
				fprintf(stderr, "%s:%d -ENOMEM\n", __func__, __LINE__);
				radio->num_unassoc_sta = 0;
				return false;
			}
		}

		for (i = 0; i < p->num_sta; i++) {
			unassoc_sta = radio->unassoc_stalist + i;
			memcpy(unassoc_sta->macaddr, p->sta[i].macaddr, 6);
			unassoc_sta->rssi = p->sta[i].ul_rcpi;
		}
	}

	return true;
}

static bool fill_scanres_channel_from_tlv(struct decollector_private *priv,
		struct cmdu_buff *resp, struct wifi_network_device *dev)
{
	int i, j;
	int radio_index;
	int num = 256;
	struct wifi_radio_element *radio;
	struct tlv *tv_tsp[1][16];
	struct tlv *tv_scan[256];

	if (!validate_channel_scan_report(resp, tv_tsp, tv_scan, &num)) {
		fprintf(stderr, "%s: cmdu validation: "\
				"[CHANNEL_SCAN_REPORT] failed\n", __func__);
		return false;
	}

	/* Parse Timestamp TLV */
	if (tv_tsp[0][0]) {
		struct tlv_timestamp *p =
			(struct tlv_timestamp *)tv_tsp[0][0]->data;

		if (!p->len)
			return false;

		for (i = 0; i < dev->num_radios; i++) {
			radio = dev->radiolist + i;

			if (radio->num_scanresult == 0) {
				/* TODO/revisit: Assuming single scanresult per radio */
				radio->num_scanresult = 1;
				radio->scanlist = calloc(radio->num_scanresult,
						sizeof(struct wifi_scanres_element));

				if (!radio->scanlist) {
					fprintf(stderr, "%s:%d -ENOMEM\n", __func__, __LINE__);
					radio->num_scanresult = 0;
					return false;
				}

				strncpy(radio->scanlist[0].tsp, (char *)p->timestamp, p->len);
			}
		}
	}

	/* Parse Channel Scan Result TLV */
	for (i = 0; i < num; i++) {
		int offset = 0;
		int opclass_index;
		int channel_index;
		uint8_t *tv_data = (uint8_t *)tv_scan[i]->data;
		struct tlv_channel_scan_result *p =
			(struct tlv_channel_scan_result *)tv_data;

		offset = sizeof(*p);
		radio_index = get_radio_index(dev, p->radio);
		if (radio_index == -1)
			continue;

		radio = dev->radiolist + radio_index;
		opclass_index = get_opclasss_index(radio, p->opclass);
		if (opclass_index == -1) {
			radio->scanlist[0].num_opclass_scanned++;
			radio->scanlist[0].opclass_scanlist = realloc(radio->scanlist[0].opclass_scanlist,
					radio->scanlist[0].num_opclass_scanned *
					sizeof(struct wifi_scanres_opclass_element));

			if (!radio->scanlist[0].opclass_scanlist) {
				fprintf(stderr, "%s:%d -ENOMEM\n", __func__, __LINE__);
				radio->scanlist[0].num_opclass_scanned = 0;
				goto error;
			}

			opclass_index = radio->scanlist[0].num_opclass_scanned - 1;
			memset(radio->scanlist[0].opclass_scanlist + opclass_index, 0,
					sizeof(struct wifi_scanres_opclass_element));
		}

		struct wifi_scanres_opclass_element *op =
			radio->scanlist[0].opclass_scanlist + opclass_index;

		op->opclass = p->opclass;
		channel_index = get_channel_index(op, p->channel);
		if (channel_index == -1) {
			op->num_channels_scanned++;
			op->channel_scanlist = realloc(op->channel_scanlist,
					op->num_channels_scanned *
					sizeof(struct wifi_scanres_channel_element));

			if (!op->channel_scanlist) {
				fprintf(stderr, "%s:%d -ENOMEM\n", __func__, __LINE__);
				op->num_channels_scanned = 0;
				goto error;
			}

			channel_index = op->num_channels_scanned - 1;
			memset(op->channel_scanlist + channel_index, 0,
					sizeof(struct wifi_scanres_channel_element));
		}

		struct wifi_scanres_channel_element *ch =
			op->channel_scanlist + channel_index;

		ch->channel = p->channel;

		/* todo/fixme: map these
		 * totoal_scan_duration & type
		 */
		if (p->status == CH_SCAN_STATUS_SUCCESS) {
			int time_len;

			time_len = tv_data[offset++];
			if (time_len > 0)
				memcpy(ch->tsp, &tv_data[offset], time_len);

			offset += time_len;
			ch->utilization = tv_data[offset++];
			ch->anpi = tv_data[offset++];
			ch->num_neighbors = BUF_GET_BE16(tv_data[offset]);
			offset += 2;

			if (ch->num_neighbors > 0)
				ch->nbrlist = calloc(ch->num_neighbors, sizeof(*ch->nbrlist));

			if (!ch->nbrlist) {
				offset += 5;
				ch->num_neighbors = 0;
				continue;
			}

			for (j = 0; j < ch->num_neighbors; j++) {
				int ssid_len = 0;
				int bw_len = 0;
				struct wifi_scanres_neighbor_element *nbr = ch->nbrlist + j;

				memcpy(nbr->bssid, &tv_data[offset], 6);
				offset += 6;
				ssid_len = tv_data[offset++];
				if (ssid_len > 0)
					memcpy(nbr->ssid, &tv_data[offset], ssid_len);

				offset += ssid_len;
				nbr->rssi = tv_data[offset++];
				/* TODO/revisit */
				bw_len = tv_data[offset++];		/* bandwidth len */
				nbr->bw = atoi((char *)&tv_data[offset]);
				offset += bw_len;

				if (tv_data[offset++] & CH_SCAN_RESULT_BSSLOAD_PRESENT) {
					nbr->utilization = tv_data[offset++];
					nbr->num_stations = BUF_GET_BE16(tv_data[offset]);
					offset += 2;
				}
			}
			offset += 5;
		}
	}

	return true;

error:
	return false;
}

static bool fill_curr_opclass_from_tlv(struct decollector_private *priv,
		struct cmdu_buff *resp, struct wifi_network_device *dev)
{
	int i;
	int index, offset = 0;
	int radio_index;
	struct wifi_radio_element *radio;
	struct wifi_opclass_current_element *opclass;
	struct tlv *tv[1][16];

	if (!validate_operating_channel_report(resp, tv)) {
		fprintf(stderr, "%s: cmdu validation "\
				"[OPERATING_CHANNEL_REPORT] failed\n",
				__func__);
		return false;
	}

	/* free old radio->curr_opclass */
	for (i = 0; i < dev->num_radios; i++) {
		radio = dev->radiolist + i;

		if (radio->num_curr_opclass > 0)
			free(radio->curr_opclass);

		radio->num_curr_opclass = 0;
		radio->curr_opclass = NULL;
	}

	index = 0;
	/* Parse Operating Channel Report TLV */
	while (tv[0][index] && (index < 16)) {
		uint8_t *tv_data = (uint8_t *)tv[0][index++]->data;
		uint8_t txpower;
		struct tlv_oper_channel_report *p =
			(struct tlv_oper_channel_report *)tv_data;

		radio_index = get_radio_index(dev, p->radio);
		if (radio_index == -1)
			continue;

		radio = dev->radiolist + radio_index;
		radio->num_curr_opclass = p->report.num_opclass;

		if (radio->num_curr_opclass > 0)
			radio->curr_opclass = calloc(radio->num_curr_opclass,
					sizeof(struct wifi_opclass_current_element));

		if (!radio->curr_opclass)
			return false;

		offset = 6 + 1;
		for (i = 0; i < radio->num_curr_opclass; i++) {
			opclass = radio->curr_opclass + i;
			get_timestamp(NULL, opclass->tsp);
			opclass->id = tv_data[offset++];
			opclass->channel = tv_data[offset++];
		}
		txpower = tv_data[offset++];

		for (i = 0; i < radio->num_curr_opclass; i++) {
			opclass = radio->curr_opclass + i;
			opclass->txpower = txpower;
		}
	}

	return true;
}

static int prepare_client_cap_query(struct decollector_private *priv, uint8_t *sta_mac,
					uint8_t *bssid, uint8_t *origin)
{
	int ret;
	uint16_t mid = 0;
	uint16_t resp_expected = 0xffff;
	struct cmdu_buff *req;

	req = cmdu_alloc_simple(CMDU_CLIENT_CAPABILITY_QUERY, &mid);
	if (!req) {
		fprintf(stderr, "%s: -ENOMEM\n", __func__);
		return -1;
	}

	memcpy(req->origin, origin, 6);
	resp_expected = CMDU_CLIENT_CAPABILITY_REPORT;

	/* Client Info TLV */
	ret = decollector_gen_client_info(priv, req, sta_mac, bssid);
	if (ret)
		goto error;

	cmdu_put_eom(req);
	ret = decollector_send_cmdu_request(priv, req, resp_expected);
	cmdu_free(req);

	return 0;

error:
	cmdu_free(req);

	return -1;
}

uint8_t *find_ie(uint8_t *ies, size_t len, uint8_t eid)
{
	uint8_t *end;

	if (!ies || len < 2)
		return NULL;

	end = ies + len;
	while (end - ies > 1) {
		if (ies + ies[1] + 2 > end)
			return NULL;

		if (ies[0] == eid)
			return ies;

		ies += ies[1] + 2;
	}

	return NULL;
}

uint8_t *find_ie_ext(uint8_t *ies, size_t len, uint8_t ext_id)
{
        uint8_t *end;

        if (!ies || len < 2)
                return NULL;

        end = ies + len;
        while (end - ies > 1) {
                if (ies + ies[1] + 2 > end)
                        return NULL;

                if (ies[0] == 0xff && ies[2] == ext_id)
                        return ies;

                ies += ies[1] + 2;
        }

        return NULL;
}

static void parse_iecap(struct wifi_caps_element *caps, uint8_t *ie, uint8_t len, uint8_t valid_ie)
{
	int i;

	if (valid_ie == HT_CAP_VALID) {
		int l;
		int max_mcs = -1;
		int nss = 0;
		int octet = 0;

		/* HT cap info (2) +
		 * A-MPDU param (1)
		 */
		int offset = 2 + 1;
		uint8_t *supp_mcs = ie + offset;

		caps->valid |= HT_CAP_VALID;

		for (l = 0; l < 76; l++) {
			if (l && !(l % 8))
				octet++;

			if (!!(supp_mcs[octet] & (1 << (l % 8))))
				max_mcs++;
		}

		if (max_mcs > 0 && max_mcs % 8 == 0)
			max_mcs--;

		nss = (max_mcs + 1) / 8;
		caps->ht |= ((nss - 1) & 0x03) << 6;
		caps->ht |= ((nss - 1) & 0x03) << 4;
		caps->ht |= (!!(ie[0] & BIT(5)) ? 1 : 0) << 3;
		caps->ht |= (!!(ie[0] & BIT(6)) ? 1 : 0) << 2;
		caps->ht |= (!!(ie[0] & BIT(1)) ? 1 : 0) << 1;
	}

	if (valid_ie == VHT_CAP_VALID) {
		int offset = 0;
		int vht_chw = 0;
		int vht_nss = 0;

		vht_chw = (ie[0] & 0xc);
		vht_nss = (ie[3] & 0xc);

		caps->vht[4] |= (!!(ie[0] & BIT(5)) ? 1 : 0) << 1;
		caps->vht[4] |= (!!(ie[0] & BIT(6)) ? 1 : 0) << 0;

		if (vht_chw >= 2 || vht_nss >= 3)
			caps->vht[5] |= 1 << 7;

		if (vht_chw >= 1)
			caps->vht[5] |= 1 << 6;

		caps->vht[5] |= (!!(ie[1] & BIT(3)) ? 1 : 0) << 5;
		caps->vht[5] |= (!!(ie[2] & BIT(3)) ? 1 : 0) << 4;

		/* VHT caps info */
		offset += 4;
		for (i = 0; i < 2; i++) {
			uint8_t *mcs_map = &ie[offset + i*4];
			int octet = 0;
			int nss = 0;
			int l;

			for (l = 0; l < 16; l += 2) {
				uint8_t supp_mcs_mask = 0;

				if (l && !(l % 8))
					octet++;

				supp_mcs_mask = mcs_map[octet] & (0x3 << (l % 8));
				supp_mcs_mask >>= (l % 8);
				if (supp_mcs_mask == 3)
					break;

				nss++;
			}

			nss = nss ? (nss-1) : nss;
			if (i == 0) {
				memcpy(&caps->vht[2], mcs_map, 2);
				caps->vht[4] |= (nss & 0x07) << 2;
			} else {
				memcpy(&caps->vht[0], mcs_map, 2);
				caps->vht[4] |= (nss & 0x07) << 5;
			}
		}
	}

	if (valid_ie == HE_CAP_VALID) {
		int offset = 0;
		uint8_t supp_chwidth = 0x00;
		bool b2, b3;
		int mcs_maplen = 4;
		uint8_t *mcs_map;
		int nss = 0;
		int index = 0;

		caps->valid |= HE_CAP_VALID;
		/* byte_mac (6) */
		offset += 6;
		supp_chwidth = (ie[offset] & 0xf7) >> 1;
		b2 = !!(supp_chwidth & BIT(2));
		b3 = !!(supp_chwidth & BIT(3));

		/* get max MCS and NSS */
		if (b2)
			mcs_maplen += 4;
		if (b3)
			mcs_maplen += 4;

		/* byte_phy (11) */
		offset += 11;
		for (i = 0; i < mcs_maplen; i += 2) {
			int octet = 0;
			int l;

			nss = 0;
			mcs_map = &ie[offset + i];

			for (l = 0; l < 16; l += 2) {
				uint8_t supp_mcs_mask = 0;

				if (l && !(l % 8))
					octet++;

				supp_mcs_mask = mcs_map[octet] & (0x3 << (l % 8));
				supp_mcs_mask >>= (l % 8);
				if (supp_mcs_mask == 3)
					break;

				nss++;
			}
		}

		nss = nss ? (nss-1) : nss;
		caps->he[0] = mcs_maplen;
		memcpy(&caps->he[1], &ie[offset], mcs_maplen);
		index = mcs_maplen + 1;
		caps->he[index] |= (nss & 0x07) << 5;
		caps->he[index] |= (nss & 0x07) << 2;
		caps->he[index] |= (!!(ie[6] & BIT(4)) ? 1 : 0) << 1;
		caps->he[index] |= (!!(ie[6] & BIT(4)) ? 1 : 0) << 0;
		index++;
		caps->he[index] |= (!!(ie[9] & BIT(7)) ? 1 : 0) << 7;
		caps->he[index] |= (!!(ie[10] & BIT(1)) ? 1 : 0) << 6;
		/* TODO: Need to add mapping for followings;
		 * UL MU-MIMO capable, UL MU-MIMO + OFDMA capable,
		 * DL MU-MIMO + OFDMA capable, UL/DL OFDMA capable
		 */
	}
}

void fill_assoc_sta_caps(struct tlv_client_cap_report *p, int len,
	struct wifi_caps_element *caps)
{
	uint8_t *ht_ie, *vht_ie, *he_ie;
	uint8_t ie_len = 0;

	if (!p->frame)
		return;

	ht_ie = find_ie(p->frame, len, IE_HT_CAP);
	vht_ie = find_ie(p->frame, len, IE_VHT_CAP);
	he_ie = find_ie_ext(p->frame, len, IE_EXT_HE_CAP);

	if (ht_ie) {
		fprintf(stdout, "HT Caps..\n");
		ie_len = ht_ie[1];
		parse_iecap(caps, &ht_ie[2], ie_len, HT_CAP_VALID);
	}

	if (vht_ie) {
		fprintf(stdout, "VHT Caps..\n");
		ie_len = vht_ie[1];
		parse_iecap(caps, &vht_ie[2], ie_len, VHT_CAP_VALID);
	}

	if (he_ie) {
		fprintf(stderr, "HE caps...\n");
		ie_len = he_ie[1];
		parse_iecap(caps, &he_ie[3], ie_len - 1, HE_CAP_VALID);
	}
}

static int fill_client_capability(struct decollector_private *priv,
					struct cmdu_buff *resp,
					struct wifi_network_device *dev)
{
	struct wifi_sta_events *sv = priv->sv;
	struct wifi_assoc_events *assoc_ev = NULL;
	struct wifi_sta_element *sta = NULL;
	struct tlv *tv[3][16];

	if (!validate_client_cap_report(resp, tv)) {
		fprintf(stderr, "%s: cmdu validation "\
				"[CLIENT_CAPABILITY_REPORT] failed\n",
				__func__);
		return -1;
	}

	/* client info TLV */
	if (tv[0][0]) {
		bool is_sta_elem = false;
		int j, k, l;
		struct wifi_radio_element *radio;
		struct wifi_bss_element *bss;
		struct tlv_client_info *p =
			(struct tlv_client_info *)tv[0][0]->data;

		/* check the sta & bssid in the assoc event list */
		for (j = (sv->num_assoc_events - 1); j >= 0; j--) {
			if ((hwaddr_equal(sv->assoclist[j].macaddr, p->macaddr)) &&
					(hwaddr_equal(sv->assoclist[j].bssid, p->bssid))) {
				fprintf(stderr, "Found sta: " MACFMT " at index: %d\n",
						MAC2STR(p->macaddr), j);
				assoc_ev = sv->assoclist + j;
				break;
			}
		}

		/* check the sta & bssid in the bss-stalist */
		for (j = 0; (j < dev->num_radios) && !is_sta_elem; j++) {
			radio = dev->radiolist + j;
			for (k = 0; (k < radio->num_bss) && !is_sta_elem; k++) {
				bss = radio->bsslist + k;
				for (l = 0; (l < bss->num_stations) && !is_sta_elem; l++) {
					if (hwaddr_equal(bss->stalist[l].macaddr, p->macaddr) &&
							hwaddr_equal(bss->bssid, p->bssid)) {
						fprintf(stderr, "Fount sta: " MACFMT " in bss: " MACFMT " assoclist\n",
								MAC2STR(p->macaddr), MAC2STR(p->bssid));
						is_sta_elem = true;
						sta = bss->stalist + l;
					}
				}
			}
		}
	}

	/* Error Code TLV */
	if (tv[2][0]) {
		struct tlv_error_code *p = (struct tlv_error_code *)tv[2][0]->data;

		if (p->reason == 0x02) {
			fprintf(stderr, "STA [" MACFMT "] not associated with any BSS\n",
					MAC2STR(p->macaddr));
		} else if (p->reason == 0x03) {
			fprintf(stderr, "Client capability report unspecified failure\n");
		}
	}

	/* Client Capability Report TLV */
	if (tv[1][0]) {
		int len;
		struct tlv *t = (struct tlv *)tv[1][0];
		struct tlv_client_cap_report *p =
			(struct tlv_client_cap_report *)t->data;

		len = BUF_GET_BE16(t->len) - 1;	/* result code */
		if (p->result == 0x00) {
			if (assoc_ev)
				fill_assoc_sta_caps(p, len, &assoc_ev->caps);

			if (sta)
				fill_assoc_sta_caps(p, len, &sta->caps);
		}
	}

	return 0;
}

static int decollector_handle_topology_notification(struct decollector_private *priv,
						struct cmdu_buff *resp)
{
	int ret, index;
	struct wifi_sta_events *sv = priv->sv;
	struct wifi_assoc_events *assoc_ev = NULL;
	struct wifi_disassoc_events *disassoc_ev = NULL;
	struct tlv *tv[2][16];

	if (!validate_topology_notification(resp, tv)) {
		fprintf(stderr, "%s: cmdu validation "\
				"[TOPOLOGY_NOTIFICATION] failed\n", __func__);
		return -1;
	}

	if (tv[1][0]) {
		struct tlv_client_assoc_event *p =
			(struct tlv_client_assoc_event *)tv[1][0]->data;

		if (p->event & CLIENT_EVENT_MASK) {
			/* client association */
			sv->num_assoc_events++;
			sv->assoclist = (struct wifi_assoc_events *)realloc(sv->assoclist,
					sv->num_assoc_events * sizeof(struct wifi_assoc_events));
			if (!sv->assoclist)
				return -1;

			index = sv->num_assoc_events - 1;
			assoc_ev = sv->assoclist + index;
			memset(assoc_ev, 0, sizeof(struct wifi_assoc_events));
			get_timestamp(NULL, assoc_ev->tsp);
			memcpy(assoc_ev->macaddr, p->macaddr, 6);
			memcpy(assoc_ev->bssid, p->bssid, 6);

			/* Send client capability query */
			ret = prepare_client_cap_query(priv, p->macaddr, p->bssid, resp->origin);
			if (ret) {
				fprintf(stderr, "Error in sending client capability query\n");
				return -1;
			}

		} else {
			/* client disassociation */
			sv->num_disassoc_events++;
			sv->disassoclist = (struct wifi_disassoc_events *)realloc(sv->disassoclist,
					sv->num_disassoc_events * sizeof(struct wifi_disassoc_events));
			if (!sv->disassoclist)
				return -1;

			index = sv->num_disassoc_events - 1;
			disassoc_ev = sv->disassoclist + index;
			memset(disassoc_ev, 0, sizeof(struct wifi_disassoc_events));
			get_timestamp(NULL, disassoc_ev->tsp);
			memcpy(disassoc_ev->sta_data.macaddr, p->macaddr, 6);
			memcpy(disassoc_ev->bssid, p->bssid, 6);

		}
	}

	return 0;
}

static int fill_disassoc_stats(struct decollector_private *priv, struct cmdu_buff *resp)
{
	int i;
	struct wifi_sta_events *sv = priv->sv;
	struct wifi_disassoc_events *disassoc_ev = NULL;
	struct tlv *tv[3][16];

	if (!validate_disassoc_stats(resp, tv)) {
		fprintf(stderr, "%s: cmdu validation "\
				"[CLIENT_DISASSOCIATION] failed\n", __func__);
		return -1;
	}

	if (tv[0][0]) {
		struct tlv_sta_mac *p =
			(struct tlv_sta_mac *)tv[0][0]->data;

		for (i = (sv->num_disassoc_events - 1); i >= 0; i--) {
			if (hwaddr_equal(sv->disassoclist[i].sta_data.macaddr,
						p->macaddr)) {
				fprintf(stderr, "Found sta: " MACFMT " at index: %d\n",
						MAC2STR(p->macaddr), i);
				disassoc_ev = sv->disassoclist + i;
				break;
			}
		}
	}

	if (tv[1][0]) {
		struct tlv_reason_code *p =
			(struct tlv_reason_code *)tv[1][0]->data;

		if (disassoc_ev)
			disassoc_ev->reason_code = p->code;
	}

	if (tv[2][0]) {
		struct tlv_assoc_sta_traffic_stats *p =
			(struct tlv_assoc_sta_traffic_stats *)tv[2][0]->data;

		if (disassoc_ev) {
			disassoc_ev->sta_data.tx_bytes = BUF_GET_BE32(p->tx_bytes);
			disassoc_ev->sta_data.rx_bytes = BUF_GET_BE32(p->rx_bytes);
			disassoc_ev->sta_data.tx_pkts = BUF_GET_BE32(p->tx_packets);
			disassoc_ev->sta_data.rx_pkts = BUF_GET_BE32(p->rx_packets);
			disassoc_ev->sta_data.tx_errors = BUF_GET_BE32(p->tx_err_packets);
			disassoc_ev->sta_data.rx_errors = BUF_GET_BE32(p->rx_err_packets);
			disassoc_ev->sta_data.rtx_pkts = BUF_GET_BE32(p->rtx_packets);
		}
	}

	return 0;
}

/* Check the assoclist on each device,
 * send the client capability query for each STA.
 */
void get_assoc_sta_caps(struct decollector_private *priv)
{
	int i, j, k, l;
	struct wifi_network *network = &priv->dm->network;

	for (i = 0; i < network->num_devices; i++) {
		struct wifi_network_device *dev = &network->device[i];

		for (j = 0; j < dev->num_radios; j++) {
			struct wifi_radio_element *radio = dev->radiolist + j;

			for (k = 0; k < radio->num_bss; k++) {
				struct wifi_bss_element *bss = radio->bsslist + k;

				for (l = 0; l < bss->num_stations; l++) {
					struct wifi_sta_element *sta = bss->stalist + l;

					fprintf(stderr, "client capability query:\n\tsta: "
							MACFMT "\n\tbssid: " MACFMT "\n",
							MAC2STR(sta->macaddr), MAC2STR(bss->bssid));

					/* send client capability query to associated clients */
					prepare_client_cap_query(priv, sta->macaddr,
							bss->bssid, dev->macaddr);
				}
			}
		}
	}
}

int prepare_1905_query(struct decollector_private *priv, uint16_t cmdu_type,
			struct wifi_network_device *dev)
{
	int ret;
	uint16_t resp_expected = 0xffff;
	struct cmdu_buff *req;
	uint16_t mid = 0;
	uint32_t i, j;
	struct wifi_radio_element *radio;
	struct wifi_bss_element *bss;

	req = cmdu_alloc_simple(cmdu_type, &mid);
	if (!req) {
		fprintf(stderr, "%s: -ENOMEM\n", __func__);
		return -1;
	}

	fprintf(stderr,
		"prepare 1905 Query: >>>>> type = %02x >>>>>>>>>>>>>>>>\n",
		cmdu_type);

	/* XXX: in decollector_send_cmdu_request(), "dst_macaddr" is set from cmdu->origin */
	memcpy(req->origin, dev->macaddr, 6);

	switch (cmdu_type) {
	case CMDU_TYPE_TOPOLOGY_QUERY:
		{
			/* Mutli-AP Profile TLV */
			ret = decollector_gen_map_profile(priv, req,
							  MULTIAP_PROFILE_2);
			if (ret)
				goto error;

			resp_expected = CMDU_TYPE_TOPOLOGY_RESPONSE;
			break;
		}

	case CMDU_AP_CAPABILITY_QUERY:
		{
			resp_expected = CMDU_AP_CAPABILITY_REPORT;
			break;
		}

	case CMDU_BACKHAUL_STA_CAPABILITY_QUERY:
		{
			resp_expected = CMDU_BACKHAUL_STA_CAPABILITY_REPORT;
			break;
		}

	case CMDU_CHANNEL_SELECTION_REQ:
		{
			for (i = 0; i < dev->num_radios; i++) {
				uint8_t power_limit = 30;

				radio = dev->radiolist + i;

				/* Channel Preference TLV */
				ret = decollector_gen_channel_pref(priv, req, radio);
				if (ret)
					goto error;

				/* Transmit Power Limit TLV */
				ret = decollector_gen_txpower_limit(priv, req,
						radio->macaddr, power_limit);
				if (ret)
					goto error;
			}

			resp_expected = CMDU_CHANNEL_SELECTION_RESPONSE;
			break;
		}

	case CMDU_AP_METRICS_QUERY:
		{
			int total_bss = 0;
			int c_index = 0;
			uint8_t *bsslist = NULL;

			for (i = 0; i < dev->num_radios; i++) {
				radio = dev->radiolist + i;

				if (dev->map_profile == MAP_PROFILE_2) {
					/* Radio Identifier TLV */
					ret = decollector_gen_ap_radio_identifier(priv,
							req, radio->macaddr);
					if (ret)
						goto error;
				}
				total_bss += radio->num_bss;
			}

			if (total_bss <= 0) {
				resp_expected = CMDU_AP_METRICS_RESPONSE;
				break;
			}

			bsslist = calloc(total_bss, 6 * sizeof(uint8_t));
			if (!bsslist) {
				fprintf(stderr, "%s:%d -ENOMEM\n", __func__, __LINE__);
				goto error;
			}

			for (i = 0; i < dev->num_radios; i++) {
				radio = dev->radiolist + i;

				for (j = 0; j < radio->num_bss; j++) {
					bss = radio->bsslist + j;
					memcpy(&bsslist[c_index * 6], bss->bssid, 6);
					c_index++;
				}
			}

			/* AP Metrics TLV */
			ret = decollector_gen_ap_metric_query(priv, req, total_bss, bsslist);
			free(bsslist);

			if (ret)
				goto error;

			resp_expected = CMDU_AP_METRICS_RESPONSE;
			break;
		}

	case CMDU_UNASSOC_STA_LINK_METRIC_QUERY:
		// TODO: Need to create the CMDU, unassoc sta list ??
		break;

	case CMDU_CHANNEL_SCAN_REQUEST:
		{
			ret = decollector_gen_channel_scan_request(priv, req, dev);
			if (ret)
				goto error;

			resp_expected = CMDU_CHANNEL_SCAN_REPORT;
			break;
		}

	default:
		fprintf(stdout, "%s:%d cmdu %d is not handled\n", __func__, __LINE__, cmdu_type);
		break;
	}

	cmdu_put_eom(req);
	decollector_send_cmdu_request(priv, req, resp_expected);

error:
	cmdu_free(req);

	return 0;
}

int send_channel_selection_req(struct decollector_private *p)
{
        int i, ret = 0;
        struct wifi_network *net = &p->dm->network;

        for (i = 0; i < net->num_devices; i++) {
                struct wifi_network_device *dev = &net->device[i];

                ret |= prepare_1905_query(p, CMDU_CHANNEL_SELECTION_REQ, dev);
        }

        return ret;
}


void decollector_update_datamodel(struct decollector_private *p)
{
	struct wifi_data_element *dm_old;

	if (!p->dmi) {
		fprintf(stderr, "%s: No new DM - exit\n", __func__);
		return;
	}

	decollector_notify_events(p);

	fprintf(stderr, "%s: Updating DM in Collector...\n", __func__);
	dm_old = p->dm;
	p->dm = p->dmi;
	p->dm_available = 1;

	decollector_free_dm(dm_old);
	p->dmi = NULL;
	/* update controller id */
	memcpy(p->dm->network.cntlr_id, p->cntlr_id, 6);

	p->getter_running = 0;

	get_assoc_sta_caps(p);

	/* send channel selection req to each device */
	send_channel_selection_req(p);
}

int decollector_handle_map_response(struct decollector_private *p, struct cmdu_buff *resp)
{
	struct wifi_network_device *dev = NULL;
	bool is_send_next_query = false;
	int ret = 0;
	uint16_t message_type = cmdu_get_type(resp) & 0xffff;
	uint16_t mid = cmdu_get_mid(resp) & 0xffff;

	ret = cmdu_ackq_dequeue(&p->cmdu_ack_q, message_type,
				mid, resp->origin, NULL);
	if (ret) {
		fprintf(stderr, "wfadatad: drop unexpected CMDU (mid = %d)\n",
			mid);
		goto out;
	}

	dev = decollector_get_origin_dev(p, p->dmi, resp->origin);
	if (!dev) {
		fprintf(stderr, "wfadatad: drop CMDU response from unknown src\n");
		ret = -1;
		goto out;
	}


	switch (message_type) {
	case CMDU_TYPE_TOPOLOGY_RESPONSE:
		//TODO: discard response in wrong state?
		is_send_next_query = fill_topology_resp_from_tlv(p, resp, dev);

		if (is_send_next_query)
			prepare_1905_query(p, CMDU_AP_CAPABILITY_QUERY, dev);
		break;

	case CMDU_AP_CAPABILITY_REPORT:
		is_send_next_query = fill_ap_capability_from_tlv(p, resp, dev);

		if (is_send_next_query)
			prepare_1905_query(p, CMDU_AP_METRICS_QUERY, dev);
		break;

	case CMDU_AP_METRICS_RESPONSE:
		is_send_next_query = fill_apstats_from_tlv(p, resp, dev);

		if ((dev->map_profile == MAP_PROFILE_2) &&
				(is_send_next_query)) {
			prepare_1905_query(p, CMDU_BACKHAUL_STA_CAPABILITY_QUERY, dev);
		} else if (dev->map_profile == MAP_PROFILE_1)
			p->pending--;

		break;

	case CMDU_BACKHAUL_STA_CAPABILITY_REPORT:
		is_send_next_query = fill_bksta_from_tlv(p, resp, dev);

		if (is_send_next_query)
			prepare_1905_query(p, CMDU_CHANNEL_SCAN_REQUEST, dev);
		break;

	case CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE:
		fill_unassocsta_from_tlv(p, resp, dev);
		break;

	case CMDU_CHANNEL_SCAN_REPORT:
		fill_scanres_channel_from_tlv(p, resp, dev);
		p->pending--;
		break;

	default:
		fprintf(stdout, "unknown CMDU:%d", message_type);
		break;
	}


	/* update datamodel if success */
	if (!ret && p->pending <= 0)
		decollector_update_datamodel(p);

out:
	return ret;
}

int decollector_handle_operchan_report(struct decollector_private *p,
		struct cmdu_buff *resp)
{
	struct wifi_network_device *dev = NULL;

	if (p->getter_running)
		return -1;

	dev = decollector_get_origin_dev(p, p->dm, resp->origin);
	if (!dev) {
		fprintf(stderr, "decollector: drop CMDU response from unknown src\n");
		return -1;
	}

	/* fill operating channel data */
	if (fill_curr_opclass_from_tlv(p, resp, dev))
		return 0;

	return -1;
}

int decollector_handle_autoconfig_response(struct decollector_private *p,
		struct cmdu_buff *resp)
{
	bool is_cntlr = false;
	int i;
	struct tlv *tv[4][16];

	if (!validate_autoconfig_response(resp, tv)) {
		fprintf(stderr, "%s: cmdu validation "\
				"[AP_AUTOCONFIGURATION_RESPONSE] failed\n",
				__func__);
		return -1;
	}

	if (tv[0][0]->data[0] > 0) {
		for (i = 0; i < tv[0][0]->data[0]; i++) {
			if (tv[0][0]->data[i+1] ==
					SUPPORTED_SERVICE_MULTIAP_CONTROLLER) {
				is_cntlr = true;
				break;
			}
		}
	}

	if (!is_cntlr) {
		fprintf(stderr, "Autoconfig response was not from controller\n");
		return -1;
	}

	fprintf(stdout, "controller id: " MACFMT "\n", MAC2STR(resp->origin));

	/* store controller id */
	memcpy(p->cntlr_id, resp->origin, 6);

	return 0;
}

int decollector_handle_map_sta_event(struct decollector_private *p, struct cmdu_buff *cmdu)
{
	int ret = 0;
	int i;
	bool is_known_src = false;
	struct wifi_network_device *dev = NULL;
	uint16_t message_type = cmdu_get_type(cmdu);

	if (!p->dm)
		return -1;

	for (i = 0; i < p->dm->network.num_devices; i++) {
		if (hwaddr_equal(p->dm->network.device[i].macaddr, cmdu->origin)) {
			is_known_src = true;
			dev = &p->dm->network.device[i];
			break;
		}
	}

	if (!is_known_src)
		return -1;

	switch (message_type) {
	case CMDU_TYPE_TOPOLOGY_NOTIFICATION:
		ret = decollector_handle_topology_notification(p, cmdu);
		break;

	case CMDU_CLIENT_DISASSOCIATION_STATS:
		ret = fill_disassoc_stats(p, cmdu);
		break;

	case CMDU_CLIENT_CAPABILITY_REPORT:
		ret = fill_client_capability(p, cmdu, dev);
		break;
	}

	return ret;
}

int prepare_autoconfig_search(struct decollector_private *priv, uint8_t band)
{
	int ret;
	uint16_t mid = 0;
	uint16_t resp_expected;
	struct cmdu_buff *req;

	req = cmdu_alloc_simple(CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH, &mid);
	if (!req) {
		fprintf(stderr, "%s: -ENOMEM\n", __func__);
		return -1;
	}

	CMDU_SET_RELAY_MCAST(req->cdata);
	memcpy(req->origin, "\x01\x80\xC2\x00\x00\x13", 6);
	resp_expected = CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE;

	/* 1905.1 AL MAC address type TLV */
	ret = decollector_gen_aladdr(priv, req);
	if (ret)
		goto error;

	/* SearchedRole TLV */
	ret = decollector_gen_searched_role(priv, req,
			IEEE80211_ROLE_REGISTRAR);
	if (ret)
		goto error;

	/* AutoconfigFreqBand TLV */
	ret = decollector_gen_autoconfig_band(priv, req, band);
	if (ret)
		goto error;

	/* SupportedService TLV */
	ret = decollector_gen_supported_service(priv, req,
			SUPPORTED_SERVICE_MULTIAP_AGENT);
	if (ret)
		goto error;

	/* SearchedService TLV */
	ret = decollector_gen_searched_service(priv, req,
			SEARCHED_SERVICE_MULTIAP_CONTROLLER);
	if (ret)
		goto error;

	if (priv->cfg.profile == MAP_PROFILE_2) {
		/* Multi-AP Profile TLV */
		ret = decollector_gen_map_profile(priv, req,
				MULTIAP_PROFILE_2);
		if (ret)
			goto error;
	}

	decollector_send_cmdu_request(priv, req, resp_expected);
	cmdu_free(req);

	return 0;
error:
	cmdu_free(req);

	return -1;
}

/* Entry point for collecting a node's wifi dataelements
 * mark the node as valid/available node
 */
int decollector_collect_node(struct decollector_private *p, struct wifi_network_device *dev)
{
	int ret;

	dev->is_valid_node = true;
	ret = prepare_1905_query(p, CMDU_TYPE_TOPOLOGY_QUERY, dev);

	return ret;
}

int decollector_handle_map_event(struct decollector_private *p, struct cmdu_buff *cmdu)
{
	int ret = 0;
	uint16_t message_type = cmdu_get_type(cmdu) & 0xffff;

	switch (message_type) {
	case CMDU_TYPE_TOPOLOGY_RESPONSE:
	case CMDU_AP_CAPABILITY_REPORT:
	case CMDU_AP_METRICS_RESPONSE:
	case CMDU_BACKHAUL_STA_CAPABILITY_REPORT:
	case CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE:
	case CMDU_CHANNEL_SCAN_REPORT:
		fprintf(stderr, "%s: priv = %p (collector = %d) cmdu = %02x\n",
				__func__, p, p->cfg.is_collector,
				message_type);

		ret = decollector_handle_map_response(p, cmdu);
		break;

	case CMDU_OPERATING_CHANNEL_REPORT:
		fprintf(stderr, "%s: priv = %p (collector = %d) cmdu = %02x\n",
				__func__, p, p->cfg.is_collector,
				message_type);

		ret = decollector_handle_operchan_report(p, cmdu);
		break;

	case CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE:
		fprintf(stderr, "%s: priv = %p (%s) cmdu = %02x\n",
			__func__, p, "collector", message_type);

		ret = decollector_handle_autoconfig_response(p, cmdu);
		break;

	case CMDU_TYPE_TOPOLOGY_NOTIFICATION:
	case CMDU_CLIENT_DISASSOCIATION_STATS:
	case CMDU_CLIENT_CAPABILITY_REPORT:
		fprintf(stderr, "%s: priv = %p (collector = %d) cmdu = %02x\n",
				__func__, p, p->cfg.is_collector,
				message_type);

		ret = decollector_handle_map_sta_event(p, cmdu);
		decollector_notify_events(p);
		break;
	default:
		break;
	}

	return ret;
}

/* Handle topology change event -
 * Reset the CMDU message queue, (re)start collecting for all the nodes.
 */
int decollector_handle_topology_event(struct decollector_private *p, const char *type)
{
	fprintf(stdout, "%s: Topology_event:%s\n", __func__, type);

	cmdu_ackq_flush(&p->cmdu_ack_q);
	decollector_free_dm(p->dmi);
	p->dmi = NULL;
	p->getter_running = 0;
	decollector_getter_run(p);

	return 0;
}

