/*
 * dump2er.c
 * WiFi Data Elements json objects dump functions
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See AUTHORS file for author and contributors.
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#include <easy/easy.h>

#include "timer.h"
#include <cmdu_ackq.h>

#include "wfa_delm.h"
#include "decollector.h"
#include "common.h"
#include "config.h"


static void dump2_network(struct decollector_private *p, struct wifi_network *net, struct blob_buf *bb);
static void dump2_network_device(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev, struct blob_buf *bb);
static void dump2_network_device_default8021q(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev, struct blob_buf *bb);
static void dump2_network_device_cacstatus(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev, struct blob_buf *bb);
static void dump2_network_device_sprule(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev, struct blob_buf *bb);
static void dump2_network_device_ieee1905security(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev, struct blob_buf *bb);
static void dump2_network_device_anticipatedchannels(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev, struct blob_buf *bb);
static void dump2_network_device_multiapdevice(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev, struct blob_buf *bb);

static void dump2_network_device_radio(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		int ri, struct blob_buf *bb);
static void dump2_network_device_radio_scanresult(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		int ri, struct blob_buf *bb);
static void dump2_network_device_radio_backhaulsta(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		int ri, struct blob_buf *bb);
static void dump2_network_device_radio_scancapability(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		int ri, struct blob_buf *bb);
static void dump2_network_device_radio_caccapability(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		int ri, struct blob_buf *bb);
static void dump2_network_device_radio_capabilities(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		int ri, struct blob_buf *bb);
static void dump2_network_device_radio_opclass(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		int ri, struct blob_buf *bb);

static void dump2_network_device_radio_bss(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		struct wifi_radio_element *radio, int bi, struct blob_buf *bb);
static void dump2_network_device_radio_bss_sta(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		struct wifi_radio_element *radio, struct wifi_bss_element *bss,
		int si, struct blob_buf *bb);
static void dump2_network_device_radio_bss_sta_multiapsta(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		struct wifi_radio_element *radio, struct wifi_bss_element *bss,
		struct wifi_sta_element *sta, struct blob_buf *bb);
static void dump2_network_device_radio_bss_sta_wifi6cap(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		struct wifi_radio_element *radio, struct wifi_bss_element *bss,
		struct wifi_sta_element *sta, struct blob_buf *bb);

static int remove_spec_char(unsigned char *out, int out_len,
		const unsigned char *src, int src_len)
{
	int i;
	int copy_len = 0;

	if (!out && !src)
		return -1;

	memset(out, 0, out_len);
	copy_len = out_len > src_len ? src_len : out_len;

	for (i = 0; i < copy_len; i++) {
		if (src[i] == '\n') {
			out[i] = '\0';
			goto out;
		} else
			out[i] = src[i];
	}

	out[i] = '\0';
out:

	return 0;
}

static void radio_make_id(struct wifi_radio_element *radio, char *id)
{
	unsigned char caps_str[32];
	unsigned char caps_str1[32];
	int str1_len = sizeof(caps_str1);
	size_t out_len;

	out_len = sizeof(caps_str);
	memset(caps_str, 0, sizeof(caps_str));
	base64_encode(radio->macaddr, sizeof(radio->macaddr), caps_str, &out_len);
	remove_spec_char(caps_str1, str1_len, caps_str, out_len);

	strncpy(id, (char *)caps_str1, 32);
	id[31] = '\0';
}

static char *caps2str(void *caps, size_t caps_len, char *str)
{
	unsigned char caps_str[32] = {0};
	unsigned char caps_str1[32] = {0};
	int str1_len = sizeof(caps_str1);
	size_t out_len = sizeof(caps_str);

	base64_encode(caps, caps_len, caps_str, &out_len);
	remove_spec_char(caps_str1, str1_len, caps_str, out_len);

	strncpy(str, (char *)caps_str1, 32);
	str[31] = '\0';
	return str;
}

static char *band_list(char *str_buf, size_t str_len, uint32_t band)
{
	char *str = str_buf;
	ssize_t len = 0;

#define ADD_BAND(b) do { \
	len = snprintf(str, str_len, "%s%s", len > 0 ? "," : "", b); \
	if (len < 0 || len >= str_len) { \
		fprintf(stderr, "snprintf fail to add band"); \
		return str_buf; \
	} \
	str += len; \
	str_len -= len; \
} while (0)

	if (band & BAND_2)
		ADD_BAND("2");
	if (band & BAND_5)
		ADD_BAND("5");
	if (band & BAND_6)
		ADD_BAND("6");
	if (band & BAND_60)
		ADD_BAND("60");

#undef ADD_BAND

	return str_buf;
}

int decollector_dump2_all(struct decollector_private *p, struct blob_buf *bb)
{
	struct wifi_data_element *dm = p->dm;
	char dstr[32] = {0};
	char tsp[32] = {0};
	void *t, *t2;
	void *arr;

	if (!p->dm_available || !dm) {
		/* Relay on refresh */
		p->getter_running = 0;
		return -1;
	}

	blobmsg_add_string(bb, "date", get_date(NULL, dstr));
	blobmsg_add_string(bb, "version", "0.2");	// dm->version
	blobmsg_add_string(bb, "description", "WFA Data Elements");
	blobmsg_add_string(bb, "TimeStamp", get_timestamp(NULL, tsp));
	blobmsg_add_string(bb, "name", "wifi");

	arr = blobmsg_open_array(bb, "data");
	t = blobmsg_open_table(bb, "");
	t2 = blobmsg_open_table(bb, "wfa-dataelements:Network");

	dump2_network(p, &dm->network, bb);

	blobmsg_close_table(bb, t2);
	blobmsg_close_table(bb, t);
	blobmsg_close_array(bb, arr);

	return 0;
}

static void dump2_network(struct decollector_private *p, struct wifi_network *net, struct blob_buf *bb)
{
	struct wifi_network_device *dev = NULL;
	struct wifi_network_ssid *ssid;
	char cntlr_id[17] = {0};
	char band_str[32] = {0};
	char tsp[32] = {0};
	void *t, *t2;
	void *arr, *arr2;
	int i;
	unsigned int available_nodes;

	/* Device.WiFi.DataElements.Network. */

	available_nodes = net->num_devices;

	for (i = 0; i < net->num_devices; i++) {
		if (!net->device[i].is_valid_node)
			available_nodes--;
	}

	blobmsg_add_string(bb, "ID", "TODO");
	blobmsg_add_string(bb, "TimeStamp", get_timestamp(NULL, tsp));
	if (!p->cfg.is_sap_mode) {
		hwaddr_ntoa(net->cntlr_id, cntlr_id);
		blobmsg_add_string(bb, "ControllerID", cntlr_id);
	}

	blobmsg_add_u32(bb, "DeviceNumberOfEntries", available_nodes);

	arr = blobmsg_open_array(bb, "MSCSDisallowedStaList");
	for (i = 0; i < 0 /* TODO */; i++) {
		t = blobmsg_open_table(bb, "");
		 ; // list of MACAddress
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);

	arr = blobmsg_open_array(bb, "SCSDisallowedStaList");
	for (i = 0; i < 0 /* TODO */; i++) {
		t = blobmsg_open_table(bb, "");
		 ; // list of MACAddress
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);

	/* Device.WiFi.DataElements.Network.SSID.{i}. */
	blobmsg_add_u32(bb, "SSIDNumberOfEntries", net->num_ssid);
	arr = blobmsg_open_array(bb, "SSID");
	list_for_each_entry(ssid, &net->ssidlist, list) {
		t2 = blobmsg_open_table(bb, "");
		blobmsg_add_string(bb, "SSID", (char *)ssid->ssid); //TODO safe ssid - non printble chars
		blobmsg_add_string(bb, "Band", band_list(band_str, sizeof(band_str), ssid->band));
		blobmsg_close_table(bb, t2);
	}
	blobmsg_close_array(bb, arr); // SSID

	/* Device.WiFi.DataElements.Network.MultiAPSteeringSummaryStats. */
	t = blobmsg_open_table(bb, "MultiAPSteeringSummaryStats");
	blobmsg_add_u64(bb, "NoCandidateAPFailures", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BlacklistAttempts", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BlacklistSuccesses", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BlacklistFailures", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BTMAttempts", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BTMSuccesses", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BTMFailures", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BTMQueryResponses", 0); //TODO StatsCounter64
	blobmsg_close_table(bb, t); // MultiAPSteeringSummaryStats

	arr2 = blobmsg_open_array(bb, "DeviceList");
	for (i = 0; i < net->num_devices; i++) {
		dev = &net->device[i];
		if (!dev->is_valid_node)
			continue;

		t2 = blobmsg_open_table(bb, "");
		dump2_network_device(p, net, dev, bb);
		blobmsg_close_table(bb, t2);
	}
	blobmsg_close_array(bb, arr2);
}

static void dump2_network_device(struct decollector_private *p, struct wifi_network *net,
		struct wifi_network_device *dev, struct blob_buf *bb)
{
	char macstr[18] = {0};
	void *arr, *t;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}. */

	hwaddr_ntoa(dev->macaddr, macstr);
	blobmsg_add_string(bb, "ID", macstr);
	blobmsg_add_string(bb, "MultiAPCapabilities", "TODO base64​(1:1)");
	blobmsg_add_u32(bb, "CollectionInterval", 0); // TODO unsignedInt
	blobmsg_add_string(bb, "ReportUnsuccessfulAssociations", "TODO bool");
	blobmsg_add_u32(bb, "MaxReportingRate", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "APMetricsReportingInterval", 0); // TODO unsignedInt​(:255)
	blobmsg_add_string(bb, "Manufacturer", "IOPSYS");
	blobmsg_add_string(bb, "SerialNumber", "TODO");
	blobmsg_add_string(bb, "ManufacturerModel", "TODO");
	blobmsg_add_string(bb, "SoftwareVersion", "TODO");
	blobmsg_add_string(bb, "ExecutionEnv", "TODO");
	blobmsg_add_string(bb, "DSCPMap", "TODO hexBinary​(64)");
	blobmsg_add_u32(bb, "MaxPrioritizationRules", 0); // TODO unsignedInt​(:255)
	blobmsg_add_string(bb, "PrioritizationSupport", "TODO bool");
	blobmsg_add_u32(bb, "MaxVIDs", 0); // TODO unsignedInt
	blobmsg_add_string(bb, "APMetricsWiFi6", "TODO bool");
	blobmsg_add_string(bb, "CountryCode", "TODO");

	arr = blobmsg_open_array(bb, "LocalSteeringDisallowedSTAList");
	for (i = 0; i < 0 /* TODO */; i++) {
		t = blobmsg_open_table(bb, "");
		 ; // list of MACAddress
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);

	arr = blobmsg_open_array(bb, "BTMSteeringDisallowedSTAList");
	for (i = 0; i < 0 /* TODO */; i++) {
		t = blobmsg_open_table(bb, "");
		 ; // list of MACAddress
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);

	blobmsg_add_string(bb, "DFSEnable", "TODO bool");
	blobmsg_add_string(bb, "ReportIndependentScans", "TODO bool");
	blobmsg_add_string(bb, "AssociatedSTAinAPMetricsWiFi6", "TODO bool");
	blobmsg_add_u32(bb, "MaxUnsuccessfulAssociationReportingRate", 0); // TODO unsignedInt
	blobmsg_add_string(bb, "STASteeringState", "TODO bool");
	blobmsg_add_string(bb, "CoordinatedCACAllowed", "TODO bool");
	blobmsg_add_string(bb, "TrafficSeparationAllowed", "TODO bool");
	blobmsg_add_string(bb, "ServicePrioritizationAllowed", "TODO bool");
	blobmsg_add_u32(bb, "RadioNumberOfEntries", dev->num_radios);
	blobmsg_add_u32(bb, "Default8021QNumberOfEntries", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "SSIDtoVIDMappingNumberOfEntries", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "CACStatusNumberOfEntries", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "IEEE1905SecurityNumberOfEntries", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "SPRuleNumberOfEntries", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "AnticipatedChannelsNumberOfEntries", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "AnticipatedChannelUsageNumberOfEntries", 0); // TODO unsignedInt

	dump2_network_device_default8021q(p, net, dev, bb);
	dump2_network_device_cacstatus(p, net, dev, bb);
	dump2_network_device_sprule(p, net, dev, bb);
	dump2_network_device_ieee1905security(p, net, dev, bb);
	dump2_network_device_anticipatedchannels(p, net, dev, bb);
	dump2_network_device_multiapdevice(p, net, dev, bb);

	arr = blobmsg_open_array(bb, "RadioList");
	for (i = 0; i < dev->num_radios; i++) {
		t = blobmsg_open_table(bb, "");
		dump2_network_device_radio(p, net, dev, i, bb);
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);
}

static void dump2_network_device_default8021q(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev, struct blob_buf *bb)
{
	struct wifi_network_ssid *ssid;
	void *arr;
	void *t;
	int i;

	/* TODO this is for device or for network ? */

	/* Device.WiFi.DataElements.Network.Device.{i}.Default8021Q.{i}. */
	arr = blobmsg_open_array(bb, "Default8021Q");
	for (i = 0; i < 1; i++) {
		t = blobmsg_open_table(bb, "");
		blobmsg_add_string(bb, "Enable", net->primary_vid ? "true" : "false");
		blobmsg_add_u32(bb, "PrimaryVID", net->primary_vid);
		blobmsg_add_u32(bb, "DefaultPCP", net->default_pcp);
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr); // Default8021Q

	/* Device.WiFi.DataElements.Network.Device.{i}.SSIDtoVIDMapping.{i}. */
	arr = blobmsg_open_array(bb, "SSIDtoVIDMapping");
	list_for_each_entry(ssid, &net->ssidlist, list) {
		t = blobmsg_open_table(bb, "");
		blobmsg_add_string(bb, "SSID", (char *)ssid->ssid);
		blobmsg_add_u32(bb, "VID", ssid->vid);
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr); // SSIDtoVIDMapping
}

static void dump2_network_device_cacstatus(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev, struct blob_buf *bb)
{
	void *arr, *arr2;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.CACStatus.{i}. */
	arr = blobmsg_open_array(bb, "CACStatus");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_string(bb, "TimeStamp", "TODO dateTime");
		blobmsg_add_u32(bb, "CACAvailableChannelNumberOfEntries", 0); // TODO unsignedInt
		blobmsg_add_u32(bb, "CACNonOccupancyChannelNumberOfEntries", 0); // TODO unsignedInt
		blobmsg_add_u32(bb, "CACActiveChannelNumberOfEntries", 0); // TODO unsignedInt

		/* Device.WiFi.DataElements.Network.Device.{i}.CACStatus.{i}.CACAvailableChannel.{i}. */
		arr2 = blobmsg_open_array(bb, "CACAvailableChannel");
		for (i = 0; i < 0 /* TODO */; i++) {
			blobmsg_add_u32(bb, "OpClass", 0); // TODO unsignedInt​(:255)
			blobmsg_add_u32(bb, "Channel", 0); // TODO unsignedInt​(:255)
			blobmsg_add_u32(bb, "Minutes", 0); // TODO unsignedInt​(:65535)
		}
		blobmsg_close_array(bb, arr2); // CACAvailableChannel

		/* Device.WiFi.DataElements.Network.Device.{i}.CACStatus.{i}.CACNonOccupancyChannel.{i}. */
		arr2 = blobmsg_open_array(bb, "CACNonOccupancyChannel");
		for (i = 0; i < 0 /* TODO */; i++) {
			blobmsg_add_u32(bb, "OpClass", 0); // TODO unsignedInt​(:255)
			blobmsg_add_u32(bb, "Channel", 0); // TODO unsignedInt​(:255)
			blobmsg_add_u32(bb, "Seconds", 0); // TODO unsignedInt​(:65535)
		}
		blobmsg_close_array(bb, arr2); // CACNonOccupancyChannel

		/* Device.WiFi.DataElements.Network.Device.{i}.CACStatus.{i}.CACActiveChannel.{i}. */
		arr2 = blobmsg_open_array(bb, "CACActiveChannel");
		for (i = 0; i < 0 /* TODO */; i++) {
			blobmsg_add_u32(bb, "OpClass", 0); // TODO unsignedInt​(:255)
			blobmsg_add_u32(bb, "Channel", 0); // TODO unsignedInt​(:255)
			blobmsg_add_u32(bb, "Countdown", 0); // TODO unsignedInt
		}
		blobmsg_close_array(bb, arr2); // CACActiveChannel
	}
	blobmsg_close_array(bb, arr); // CACStatus

#if 0
	/* TODO */
	t = blobmsg_open_table(bb, "CACStatus");
	blobmsg_add_string(bb, "TimeStamp", "2019-02-11T06:23:43.750419-08:00");
	arr = blobmsg_open_array(bb, "AvailableChannelList");
	for (int i = 0; i < 0; i++) {
		t2 = blobmsg_open_table(bb, "");
		blobmsg_add_string(bb, "OpClass", "128");
		blobmsg_add_string(bb, "Channel", "100");
		blobmsg_add_string(bb, "Minutes", "10");
		blobmsg_close_table(bb, t2);
	}
	blobmsg_close_array(bb, arr);

	arr = blobmsg_open_array(bb, "NonOccupancyChannelList");
	for (int i = 0; i < 0; i++) {
		t2 = blobmsg_open_table(bb, "");
		blobmsg_add_string(bb, "OpClass", "128");
		blobmsg_add_string(bb, "Channel", "116");
		blobmsg_add_string(bb, "Seconds", "30");
		blobmsg_close_table(bb, t2);
	}
	blobmsg_close_array(bb, arr);

	arr = blobmsg_open_array(bb, "ActiveChannelList");
	for (int i = 0; i < 0; i++) {
		t2 = blobmsg_open_table(bb, "");
		blobmsg_add_string(bb, "OpClass", "128");
		blobmsg_add_string(bb, "Channel", "112");
		blobmsg_add_string(bb, "Countdown", "30");
		blobmsg_close_table(bb, t2);
	}
	blobmsg_close_array(bb, arr);
	blobmsg_close_table(bb, t);
#endif
}

static void dump2_network_device_sprule(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev, struct blob_buf *bb)
{
	void *arr;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.SPRule.{i}. */
	arr = blobmsg_open_array(bb, "SPRule");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_u32(bb, "ID", 0); // TODO unsignedInt
		blobmsg_add_u32(bb, "Precedence", 0); // TODO unsignedInt​(:254)
		blobmsg_add_u32(bb, "Output", 0); // TODO unsignedInt​(:9)
		blobmsg_add_string(bb, "AlwaysMatch", "TODO bool");
	}
	blobmsg_close_array(bb, arr);
}

static void dump2_network_device_ieee1905security(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev, struct blob_buf *bb)
{
	void *arr;
	void *t;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.IEEE1905Security.{i}. */
	arr = blobmsg_open_array(bb, "IEEE1905Security");
	for (i = 0; i < 1; i++) {
		t = blobmsg_open_table(bb, "");
		blobmsg_add_u32(bb, "OnboardingProtocol", 0); // 0: 1905 Device.
		blobmsg_add_u32(bb, "IntegrityAlgorithm", 0); // 0: HMAC-SHA256.
		blobmsg_add_u32(bb, "EncryptionAlgorithm", 0); // 0: AES-SIV.
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr); // IEEE1905Security

}

static void dump2_network_device_anticipatedchannels(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev, struct blob_buf *bb)
{
	void *arr, *arr2;
	void *t;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.AnticipatedChannels.{i}. */
	arr = blobmsg_open_array(bb, "AnticipatedChannels");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_u32(bb, "OpClass", 0); // TODO unsignedInt​(:255)
		arr = blobmsg_open_array(bb, "ChannelList");
		for (i = 0; i < 0 /* TODO */; i++) {
			t = blobmsg_open_table(bb, "");
			 ; // list of unsignedInt​(:255)
			blobmsg_close_table(bb, t);
		}
		blobmsg_close_array(bb, arr);
	}
	blobmsg_close_array(bb, arr); // AnticipatedChannels

	/* Device.WiFi.DataElements.Network.Device.{i}.AnticipatedChannelUsage.{i}. */
	arr = blobmsg_open_array(bb, "AnticipatedChannelUsage");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_u32(bb, "OpClass", 0); // TODO unsignedInt​(:255)
		blobmsg_add_u32(bb, "Channel", 0); // TODO unsignedInt​(:255)
		//hwaddr_ntoa(usta->macaddr, macaddr);)
		blobmsg_add_string(bb, "ReferenceBSSID", "TODO MACAddress");
		blobmsg_add_u32(bb, "EntryNumberOfEntries", 0); // TODO unsignedInt

		/* Device.WiFi.DataElements.Network.Device.{i}.AnticipatedChannelUsage.{i}.Entry.{i}. */
		arr2 = blobmsg_open_array(bb, "Entry");
		for (i = 0; i < 0 /* TODO */; i++) {
			blobmsg_add_string(bb, "BurstStartTime", "TODO hexBinary​(4:4)");
			blobmsg_add_u32(bb, "BurstLength", 0); // TODO unsignedInt
			blobmsg_add_u32(bb, "Repetitions", 0); // TODO unsignedInt
			blobmsg_add_u32(bb, "BurstInterval", 0); // TODO unsignedInt
			blobmsg_add_string(bb, "RUBitmask", "TODO hexBinary​(2:10)");
			//hwaddr_ntoa(usta->macaddr, macaddr);)
			blobmsg_add_string(bb, "TransmitterIdentifier", "TODO MACAddress");
			blobmsg_add_u32(bb, "PowerLevel", 0); // TODO int
			blobmsg_add_string(bb, "ChannelUsageReason", "TODO");
		}
		blobmsg_close_array(bb, arr2); // Entry
	}

	blobmsg_close_array(bb, arr); // AnticipatedChannelUsage
}

static void dump2_network_device_multiapdevice(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev, struct blob_buf *bb)
{
	void *t, *t2, *t3;
	void *arr;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.MultiAPDevice. */
	t = blobmsg_open_table(bb, "MultiAPDevice");
	blobmsg_add_string(bb, "ManufacturerOUI", "TODO");
	blobmsg_add_string(bb, "LastContactTime", "TODO dateTime");
	blobmsg_add_string(bb, "AssocIEEE1905DeviceRef", "TODO");
	blobmsg_add_string(bb, "EasyMeshControllerOperationMode", "TODO");
	blobmsg_add_string(bb, "EasyMeshAgentOperationMode", "TODO");

	/* Device.WiFi.DataElements.Network.Device.{i}.MultiAPDevice.Backhaul. */
	t2 = blobmsg_open_table(bb, "Backhaul");
	blobmsg_add_string(bb, "LinkType", "TODO");
	//hwaddr_ntoa(usta->macaddr, macaddr);)
	blobmsg_add_string(bb, "BackhaulMACAddress", "TODO MACAddress");
	//hwaddr_ntoa(usta->macaddr, macaddr);)
	blobmsg_add_string(bb, "BackhaulDeviceID", "TODO MACAddress");
	//hwaddr_ntoa(usta->macaddr, macaddr);)
	blobmsg_add_string(bb, "MACAddress", "TODO MACAddress");
	blobmsg_add_u32(bb, "CurrentOperatingClassProfileNumberOfEntries", 0); // TODO unsignedInt

	/* Device.WiFi.DataElements.Network.Device.{i}.MultiAPDevice.Backhaul.CurrentOperatingClassProfile.{i}. */

	arr = blobmsg_open_array(bb, "CurrentOperatingClassProfile");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_u32(bb, "Class", 0); // TODO unsignedInt​(:255)
		blobmsg_add_u32(bb, "Channel", 0); // TODO unsignedInt​(:255)
		blobmsg_add_u32(bb, "TxPower", 0); // TODO int​(-127:127)
		blobmsg_add_string(bb, "TimeStamp", "TODO dateTime");
	}
	blobmsg_close_array(bb, arr); // CurrentOperatingClassProfile

	/* Device.WiFi.DataElements.Network.Device.{i}.MultiAPDevice.Backhaul.Stats. */
	t3 = blobmsg_open_table(bb, "Stats");
	blobmsg_add_u64(bb, "BytesSent", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BytesReceived", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "PacketsSent", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "PacketsReceived", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "ErrorsSent", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "ErrorsReceived", 0); //TODO StatsCounter64
	blobmsg_add_u32(bb, "LinkUtilization", 0); // TODO unsignedInt​(:100)
	blobmsg_add_u32(bb, "SignalStrength", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "LastDataDownlinkRate", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "LastDataUplinkRate", 0); // TODO unsignedInt
	blobmsg_add_string(bb, "TimeStamp", "TODO dateTime");
	blobmsg_close_table(bb, t3); // Stats

	blobmsg_close_table(bb, t2); // Backhaul
	blobmsg_close_table(bb, t); // MultiAPDevice

}

static void dump2_network_device_radio(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		int ri, struct blob_buf *bb)
{
	struct wifi_radio_element *radio = dev->radiolist + ri;
	void *arr;
	void *t;
	int i;
	char id[32];

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}. */
	radio_make_id(radio, id);
	blobmsg_add_string(bb, "ID", id);
	blobmsg_add_string(bb, "Enabled", radio->enabled ? "true" : "false");
	blobmsg_add_u32(bb, "Noise", radio->anpi);
	blobmsg_add_u32(bb, "Utilization", radio->total_utilization);
	blobmsg_add_u32(bb, "Transmit", radio->tx_utilization);
	blobmsg_add_u32(bb, "ReceiveSelf", radio->rx_utilization);
	blobmsg_add_u32(bb, "ReceiveOther", radio->other_utilization);
	blobmsg_add_string(bb, "TrafficSeparationCombinedFronthaul", "TODO bool");
	blobmsg_add_string(bb, "TrafficSeparationCombinedBackhaul", "TODO bool");
	blobmsg_add_u32(bb, "SteeringPolicy", 0); // TODO unsignedInt​(:2)
	blobmsg_add_u32(bb, "ChannelUtilizationThreshold", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "RCPISteeringThreshold", 0); // TODO unsignedInt​(:220)
	blobmsg_add_u32(bb, "STAReportingRCPIThreshold", 0); // TODO unsignedInt​(:220)
	blobmsg_add_u32(bb, "STAReportingRCPIHysteresisMarginOverride", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "ChannelUtilizationReportingThreshold", 0); // TODO unsignedInt
	blobmsg_add_string(bb, "AssociatedSTATrafficStatsInclusionPolicy", "TODO bool");
	blobmsg_add_string(bb, "AssociatedSTALinkMetricsInclusionPolicy", "TODO bool");
	blobmsg_add_string(bb, "ChipsetVendor", "TODO");
	blobmsg_add_string(bb, "APMetricsWiFi6", "TODO bool");
	blobmsg_add_u32(bb, "CurrentOperatingClassProfileNumberOfEntries", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "UnassociatedSTANumberOfEntries", radio->num_unassoc_sta);
	blobmsg_add_u32(bb, "BSSNumberOfEntries", radio->num_bss);
	blobmsg_add_u32(bb, "ScanResultNumberOfEntries", radio->num_scanresult);
	blobmsg_add_u32(bb, "DisAllowedOpClassChannelsNumberOfEntries", 0); // TODO unsignedInt

	dump2_network_device_radio_scanresult(p, net, dev, ri, bb);
	dump2_network_device_radio_backhaulsta(p, net, dev, ri, bb);
	dump2_network_device_radio_scancapability(p, net, dev, ri, bb);
	dump2_network_device_radio_caccapability(p, net, dev, ri, bb);
	dump2_network_device_radio_capabilities(p, net, dev, ri, bb);
	dump2_network_device_radio_opclass(p, net, dev, ri, bb);

	arr = blobmsg_open_array(bb, "BSSList");
	for (i = 0; i < radio->num_bss; i++) {
		t = blobmsg_open_table(bb, "");
		dump2_network_device_radio_bss(p, net, dev, radio, i, bb);
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.UnassociatedSTA.{i}. */
	arr = blobmsg_open_array(bb, "UnassociatedSTA");
	for (i = 0; i < radio->num_unassoc_sta; i++) {
		struct wifi_unassoc_sta_element *usta = radio->unassoc_stalist + i;
		char usta_macaddr[17] = {0};

		hwaddr_ntoa(usta->macaddr, usta_macaddr);
		blobmsg_add_string(bb, "MACAddress", usta_macaddr);
		blobmsg_add_u32(bb, "SignalStrength", usta->rssi);
	}
	blobmsg_close_array(bb, arr); // UnassociatedSTA

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.MultiAPRadio. */
	t = blobmsg_open_table(bb, "MultiAPRadio");
	//Unknown type  list​(1024) of unsignedInt TOKEN RadarDetections
	blobmsg_close_table(bb, t); // MultiAPRadio
}

static void dump2_network_device_radio_scanresult(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		int ri, struct blob_buf *bb)
{
	struct wifi_radio_element *radio = dev->radiolist + ri;
	void *arr, *arr2, *arr3, *arr4;
	void *t, *t2, *t3, *t4;
	int i, j, k, l;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.ScanResult.{i}. */
	arr = blobmsg_open_array(bb, "ScanResult");
	for (i = 0; i < radio->num_scanresult; i++) {
		struct wifi_scanres_element *sres = radio->scanlist + i;

		t = blobmsg_open_table(bb, "");
		blobmsg_add_string(bb, "TimeStamp", sres->tsp);
		blobmsg_add_u32(bb, "OpClassScanNumberOfEntries", sres->num_opclass_scanned);

		/* ... Radio.{i}.ScanResult.{i}.OpClassScan.{i}. */
		arr2 = blobmsg_open_array(bb, "OpClassScan");
		for (j = 0; j < sres->num_opclass_scanned ; j++) {
			struct wifi_scanres_opclass_element *op = sres->opclass_scanlist + j;

			t2 = blobmsg_open_table(bb, "");
			blobmsg_add_u32(bb, "OperatingClass", op->opclass);
			blobmsg_add_u32(bb, "ChannelScanNumberOfEntries", op->num_channels_scanned);

			/* ... ScanResult.{i}.OpClassScan.{i}.ChannelScan.{i}. */
			arr3 = blobmsg_open_array(bb, "ChannelScan");
			for (k = 0; k < op->num_channels_scanned; k++) {
				struct wifi_scanres_channel_element *ch =  op->channel_scanlist + k;

				t3 = blobmsg_open_table(bb, "");
				blobmsg_add_u32(bb, "Channel", ch->channel);
				blobmsg_add_string(bb, "TimeStamp", ch->tsp);
				blobmsg_add_u32(bb, "Utilization", ch->utilization);
				blobmsg_add_u32(bb, "Noise", ch->anpi);
				blobmsg_add_u32(bb, "NeighborBSSNumberOfEntries", ch->num_neighbors);

				/* ... OpClassScan.{i}.ChannelScan.{i}.NeighborBSS.{i}. */
				arr4 = blobmsg_open_array(bb, "NeighborBSS");
				for (l = 0; l < ch->num_neighbors; l++) {
					struct wifi_scanres_neighbor_element *nbr = ch->nbrlist + l;
					char nbr_bssid[17] = {0};

					t4 = blobmsg_open_table(bb, "");
					hwaddr_ntoa(nbr->bssid, nbr_bssid);
					blobmsg_add_string(bb, "BSSID", nbr_bssid);
					blobmsg_add_string(bb, "SSID", nbr->ssid);
					blobmsg_add_u32(bb, "SignalStrength", nbr->rssi);
					blobmsg_add_u32(bb, "ChannelBandwidth", nbr->bw);
					blobmsg_add_u32(bb, "ChannelUtilization", nbr->utilization);
					blobmsg_add_u32(bb, "StationCount", nbr->num_stations);
					blobmsg_close_table(bb, t4);
				}
				blobmsg_close_array(bb, arr4); // NeighborBSS
				blobmsg_close_table(bb, t3);
			}
			blobmsg_close_array(bb, arr3); // ChannelScan
			blobmsg_close_table(bb, t2);
		}
		blobmsg_close_array(bb, arr2); // OpClassScan
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr); // ScanResult
}

static void dump2_network_device_radio_backhaulsta(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		int ri, struct blob_buf *bb)
{
	struct wifi_radio_element *radio = dev->radiolist + ri;
	struct wifi_backhaul_element *bksta = &radio->bksta;
	char macaddr[17] = {0};
	void *t;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BackhaulSta. */
	t = blobmsg_open_table(bb, "BackhaulSta");
	hwaddr_ntoa(bksta->macaddr, macaddr);
	blobmsg_add_string(bb, "MACAddress", macaddr);
	blobmsg_close_table(bb, t); // BackhaulSta
}

static void dump2_network_device_radio_scancapability(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		int ri, struct blob_buf *bb)
{
	void *arr, *arr2;
	void *t, *t2;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.ScanCapability. */
	t = blobmsg_open_table(bb, "ScanCapability");
	blobmsg_add_string(bb, "OnBootOnly", "TODO bool");
	blobmsg_add_u32(bb, "Impact", 0); // TODO unsignedInt​(:3)
	blobmsg_add_u32(bb, "MinimumInterval", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "OpClassChannelsNumberOfEntries", 0); // TODO unsignedInt

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.ScanCapability.OpClassChannels.{i}. */
	arr = blobmsg_open_array(bb, "OpClassChannels");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_u32(bb, "OpClass", 0); // TODO unsignedInt​(:255)
		arr2 = blobmsg_open_array(bb, "ChannelList");
		for (i = 0; i < 0 /* TODO */; i++) {
			t2 = blobmsg_open_table(bb, "");
			 ; // list of unsignedInt​(:255)
			blobmsg_close_table(bb, t2);
		}
		blobmsg_close_array(bb, arr2);
	}
	blobmsg_close_array(bb, arr); // OpClassChannels
	blobmsg_close_table(bb, t); // ScanCapability
}

static void dump2_network_device_radio_caccapability(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		int ri, struct blob_buf *bb)
{
	void *arr, *arr2, *arr3;
	void *t, *t2;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.CACCapability. */
	t = blobmsg_open_table(bb, "CACCapability");
	blobmsg_add_u32(bb, "CACMethodNumberOfEntries", 0); // TODO unsignedInt

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.CACCapability.CACMethod.{i}. */
	arr = blobmsg_open_array(bb, "CACMethod");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_u32(bb, "Method", 0); // TODO unsignedInt
		blobmsg_add_u32(bb, "NumberOfSeconds", 0); // TODO unsignedInt
		blobmsg_add_u32(bb, "OpClassChannelsNumberOfEntries", 0); // TODO unsignedInt

		/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.CACCapability.CACMethod.{i}.OpClassChannels.{i}. */
		arr2 = blobmsg_open_array(bb, "OpClassChannels");
		for (i = 0; i < 0 /* TODO */; i++) {
			blobmsg_add_u32(bb, "OpClass", 0); // TODO unsignedInt​(:255)
			arr3 = blobmsg_open_array(bb, "ChannelList");
			for (i = 0; i < 0 /* TODO */; i++) {
				t2 = blobmsg_open_table(bb, "");
				 ; // list of unsignedInt​(:255)
				blobmsg_close_table(bb, t2);
			}
			blobmsg_close_array(bb, arr3);
		}
		blobmsg_close_array(bb, arr2); // OpClassChannels
	}
	blobmsg_close_array(bb, arr); // CACMethod
	blobmsg_close_table(bb, t); // CACCapability
}

static void dump2_network_device_radio_capabilities(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		int ri, struct blob_buf *bb)
{
	void *arr;
	void *t, *t2;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.Capabilities. */
	t = blobmsg_open_table(bb, "Capabilities");

	blobmsg_add_string(bb, "HTCapabilities", "TODO");
	blobmsg_add_string(bb, "VHTCapabilities", "TODO");
	blobmsg_add_string(bb, "HECapabilities", "TODO");
	blobmsg_add_u32(bb, "CapableOperatingClassProfileNumberOfEntries", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "AKMFrontHaulNumberOfEntries", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "AKMBackhaulNumberOfEntries", 0); // TODO unsignedInt

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.Capabilities.WiFi6APRole. */
	t2 = blobmsg_open_table(bb, "WiFi6APRole");
	blobmsg_add_string(bb, "HE160", "TODO bool");
	blobmsg_add_string(bb, "HE8080", "TODO bool");
	blobmsg_add_string(bb, "MCSNSS", "TODO base64​(4:12)");
	blobmsg_add_string(bb, "SUBeamformer", "TODO bool");
	blobmsg_add_string(bb, "SUBeamformee", "TODO bool");
	blobmsg_add_string(bb, "MUBeamformer", "TODO bool");
	blobmsg_add_string(bb, "Beamformee80orLess", "TODO bool");
	blobmsg_add_string(bb, "BeamformeeAbove80", "TODO bool");
	blobmsg_add_string(bb, "ULMUMIMO", "TODO bool");
	blobmsg_add_string(bb, "ULOFDMA", "TODO bool");
	blobmsg_add_u32(bb, "MaxDLMUMIMO", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "MaxULMUMIMO", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "MaxDLOFDMA", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "MaxULOFDMA", 0); // TODO unsignedInt​(:255)
	blobmsg_add_string(bb, "RTS", "TODO bool");
	blobmsg_add_string(bb, "MURTS", "TODO bool");
	blobmsg_add_string(bb, "MultiBSSID", "TODO bool");
	blobmsg_add_string(bb, "MUEDCA", "TODO bool");
	blobmsg_add_string(bb, "TWTRequestor", "TODO bool");
	blobmsg_add_string(bb, "TWTResponder", "TODO bool");
	blobmsg_add_string(bb, "SpatialReuse", "TODO bool");
	blobmsg_add_string(bb, "AnticipatedChannelUsage", "TODO bool");
	blobmsg_close_table(bb, t2); // WiFi6APRole

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.Capabilities.WiFi6bSTARole. */
	t2 = blobmsg_open_table(bb, "WiFi6bSTARole");
	blobmsg_add_string(bb, "HE160", "TODO bool");
	blobmsg_add_string(bb, "HE8080", "TODO bool");
	blobmsg_add_string(bb, "MCSNSS", "TODO base64​(4:12)");
	blobmsg_add_string(bb, "SUBeamformer", "TODO bool");
	blobmsg_add_string(bb, "SUBeamformee", "TODO bool");
	blobmsg_add_string(bb, "MUBeamformer", "TODO bool");
	blobmsg_add_string(bb, "Beamformee80orLess", "TODO bool");
	blobmsg_add_string(bb, "BeamformeeAbove80", "TODO bool");
	blobmsg_add_string(bb, "ULMUMIMO", "TODO bool");
	blobmsg_add_string(bb, "ULOFDMA", "TODO bool");
	blobmsg_add_u32(bb, "MaxDLMUMIMO", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "MaxULMUMIMO", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "MaxDLOFDMA", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "MaxULOFDMA", 0); // TODO unsignedInt​(:255)
	blobmsg_add_string(bb, "RTS", "TODO bool");
	blobmsg_add_string(bb, "MURTS", "TODO bool");
	blobmsg_add_string(bb, "MultiBSSID", "TODO bool");
	blobmsg_add_string(bb, "MUEDCA", "TODO bool");
	blobmsg_add_string(bb, "TWTRequestor", "TODO bool");
	blobmsg_add_string(bb, "TWTResponder", "TODO bool");
	blobmsg_add_string(bb, "SpatialReuse", "TODO bool");
	blobmsg_add_string(bb, "AnticipatedChannelUsage", "TODO bool");
	blobmsg_close_table(bb, t2); // WiFi6bSTARole

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.Capabilities.AKMFrontHaul.{i}. */
	arr = blobmsg_open_array(bb, "AKMFrontHaul");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_string(bb, "OUI", "TODO base64");
		blobmsg_add_u32(bb, "Type", 0); // TODO unsignedInt​(:255)
	}
	blobmsg_close_array(bb, arr); // AKMFrontHaul

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.Capabilities.AKMBackhaul.{i}. */
	arr = blobmsg_open_array(bb, "AKMBackhaul");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_string(bb, "OUI", "TODO base64");
		blobmsg_add_u32(bb, "Type", 0); // TODO unsignedInt​(:255)
	}
	blobmsg_close_array(bb, arr); // AKMBackhaul

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.Capabilities.CapableOperatingClassProfile.{i}. */
	arr = blobmsg_open_array(bb, "CapableOperatingClassProfile");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_u32(bb, "Class", 0); // TODO unsignedInt​(:255)
		blobmsg_add_u32(bb, "MaxTxPower", 0); // TODO int​(-127:127)
		arr = blobmsg_open_array(bb, "NonOperable");
		for (i = 0; i < 0 /* TODO */; i++) {
			t = blobmsg_open_table(bb, "");
			 ; // list of unsignedInt​(:255)
			blobmsg_close_table(bb, t);
		}
		blobmsg_close_array(bb, arr);
		blobmsg_add_u32(bb, "NumberOfNonOperChan", 0); // TODO unsignedInt
	}
	blobmsg_close_array(bb, arr); // CapableOperatingClassProfile

	blobmsg_close_table(bb, t); // Capabilities
}

static void dump2_network_device_radio_opclass(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		int ri, struct blob_buf *bb)
{
	void *arr, *arr2;
	void *t, *t2;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.CurrentOperatingClassProfile.{i}. */
	arr = blobmsg_open_array(bb, "CurrentOperatingClassProfile");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_u32(bb, "Class", 0); // TODO unsignedInt​(:255)
		blobmsg_add_u32(bb, "Channel", 0); // TODO unsignedInt​(:255)
		blobmsg_add_u32(bb, "TxPower", 0); // TODO int​(-127:127)
		blobmsg_add_u32(bb, "TransmitPowerLimit", 0); // TODO int​(-128:127)
		blobmsg_add_string(bb, "TimeStamp", "TODO");
	}
	blobmsg_close_array(bb, arr); // CurrentOperatingClassProfile

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.DisAllowedOpClassChannels.{i}. */
	arr = blobmsg_open_array(bb, "DisAllowedOpClassChannels");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_string(bb, "Enable", "TODO bool");
		blobmsg_add_u32(bb, "OpClass", 0); // TODO unsignedInt​(:255)
		arr2 = blobmsg_open_array(bb, "ChannelList");
		for (i = 0; i < 0 /* TODO */; i++) {
			t2 = blobmsg_open_table(bb, "");
			 ; // list of unsignedInt​(:255)
			blobmsg_close_table(bb, t2);
		}
		blobmsg_close_array(bb, arr2);
	}
	blobmsg_close_array(bb, arr); // DisAllowedOpClassChannels

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.SpatialReuse. */
	t = blobmsg_open_table(bb, "SpatialReuse");
	blobmsg_add_u32(bb, "PartialBSSColor", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "BSSColor", 0); // TODO unsignedInt
	blobmsg_add_string(bb, "HESIGASpatialReuseValue15Allowed", "TODO bool");
	blobmsg_add_string(bb, "SRGInformationValid", "TODO bool");
	blobmsg_add_string(bb, "NonSRGOffsetValid", "TODO bool");
	blobmsg_add_string(bb, "PSRDisallowed", "TODO bool");
	blobmsg_add_u32(bb, "NonSRGOBSSPDMaxOffset", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "SRGOBSSPDMinOffset", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "SRGOBSSPDMaxOffset", 0); // TODO unsignedInt
	blobmsg_add_string(bb, "SRGBSSColorBitmap", "TODO hexBinary​(8:8)");
	blobmsg_add_string(bb, "SRGPartialBSSIDBitmap", "TODO hexBinary​(8:8)");
	blobmsg_add_string(bb, "NeighborBSSColorInUseBitmap", "TODO hexBinary​(8:8)");
	blobmsg_close_table(bb, t); // SpatialReuse
}

static void dump2_network_device_radio_bss(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		struct wifi_radio_element *radio, int bi, struct blob_buf *bb)
{
	struct wifi_bss_element *bss = radio->bsslist + bi;
	char bssid[17] = {0};
	void *arr;
	void *t;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}. */

	//hwaddr_ntoa(usta->macaddr, macaddr);)
	hwaddr_ntoa(bss->bssid, bssid);
	blobmsg_add_string(bb, "BSSID", bssid);
	blobmsg_add_string(bb, "SSID", (char *)bss->ssid);
	blobmsg_add_string(bb, "Enabled", bss->enabled ? "true" : "false");
	blobmsg_add_u32(bb, "LastChange", bss->uptime);
	blobmsg_add_string(bb, "TimeStamp", bss->tsp);
	blobmsg_add_u64(bb, "UnicastBytesSent", bss->tx_ucast_bytes);
	blobmsg_add_u64(bb, "UnicastBytesReceived", bss->rx_ucast_bytes);
	blobmsg_add_u64(bb, "MulticastBytesSent", bss->tx_mcast_bytes);
	blobmsg_add_u64(bb, "MulticastBytesReceived", bss->rx_mcast_bytes);
	blobmsg_add_u64(bb, "BroadcastBytesSent", bss->tx_bcast_bytes);
	blobmsg_add_u64(bb, "BroadcastBytesReceived", bss->rx_bcast_bytes);
	blobmsg_add_u32(bb, "ByteCounterUnits", 0); //TODO
	blobmsg_add_string(bb, "Profile1bSTAsDisallowed", "TODO bool");
	blobmsg_add_string(bb, "Profile2bSTAsDisallowed", "TODO bool");
	blobmsg_add_u32(bb, "AssociationAllowanceStatus", 0); // TODO unsignedInt​(0:1)
	blobmsg_add_string(bb, "EstServiceParametersBE", "TODO base64​(0:3)");
	blobmsg_add_string(bb, "EstServiceParametersBK", "TODO base64​(0:3)");
	blobmsg_add_string(bb, "EstServiceParametersVI", "TODO base64​(0:3)");
	blobmsg_add_string(bb, "EstServiceParametersVO", "TODO base64​(0:3)");
	blobmsg_add_string(bb, "BackhaulUse", "TODO bool");
	blobmsg_add_string(bb, "FronthaulUse", "TODO bool");
	blobmsg_add_string(bb, "R1disallowed", "TODO bool");
	blobmsg_add_string(bb, "R2disallowed", "TODO bool");
	blobmsg_add_string(bb, "MultiBSSID", "TODO bool");
	blobmsg_add_string(bb, "TransmittedBSSID", "TODO bool");

	arr = blobmsg_open_array(bb, "FronthaulAKMsAllowed");
	for (i = 0; i < 0 /* TODO */; i++) {
		t = blobmsg_open_table(bb, "");
		 ; // list of string
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);

	arr = blobmsg_open_array(bb, "BackhaulAKMsAllowed");
	for (i = 0; i < 0 /* TODO */; i++) {
		t = blobmsg_open_table(bb, "");
		 ; // list of string
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);

	blobmsg_add_u32(bb, "STANumberOfEntries", bss->num_stations);
	blobmsg_add_u32(bb, "QMDescriptorNumberOfEntries", 0); // TODO unsignedInt

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.QMDescriptor.{i}. */
	arr = blobmsg_open_array(bb, "QMDescriptor");
	for (i = 0; i < 0 /* TODO */; i++) {
		//hwaddr_ntoa(usta->macaddr, macaddr);)
		blobmsg_add_string(bb, "ClientMAC", "TODO MACAddress");
		blobmsg_add_string(bb, "DescriptorElement", "TODO hexBinary");
	}
	blobmsg_close_array(bb, arr); // QMDescriptor

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.MultiAPSteering. */
	t = blobmsg_open_table(bb, "MultiAPSteering");
	blobmsg_add_u64(bb, "BlacklistAttempts", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BTMAttempts", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BTMQueryResponses", 0); //TODO StatsCounter64
	blobmsg_close_table(bb, t); // MultiAPSteering

	arr = blobmsg_open_array(bb, "STAList");
	for (i = 0; i < bss->num_stations; i++) {
		t = blobmsg_open_table(bb, "");
		dump2_network_device_radio_bss_sta(p, net, dev, radio, bss, i, bb);
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);
}

static void dump2_network_device_radio_bss_sta(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		struct wifi_radio_element *radio, struct wifi_bss_element *bss,
		int si, struct blob_buf *bb)
{
	struct wifi_sta_element *sta = bss->stalist + si;
	char macaddr[17] = {0};
	char caps_str[32] = {0};
	void *arr;
	void *t;
	int i;
	int he_caps_len;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.STA.{i}. */
	hwaddr_ntoa(sta->macaddr, macaddr);
	blobmsg_add_string(bb, "MACAddress", macaddr);
	blobmsg_add_string(bb, "TimeStamp", sta->tsp);

	blobmsg_add_string(bb, "HTCapabilities", caps2str(&sta->caps.ht, sizeof(sta->caps.ht), caps_str));
	blobmsg_add_string(bb, "VHTCapabilities", caps2str(&sta->caps.vht, sizeof(sta->caps.vht), caps_str));
	if (sta->caps.valid & HE_CAP_VALID)
		he_caps_len = 2 + sta->caps.he[0];
	else
		he_caps_len = 4; /* pass the usp-validator range test */
	blobmsg_add_string(bb, "HECapabilities", caps2str(&sta->caps.he[1], he_caps_len, caps_str));

	blobmsg_add_string(bb, "ClientCapabilities", "TODO base64");
	blobmsg_add_u32(bb, "LastDataDownlinkRate", sta->dl_rate);
	blobmsg_add_u32(bb, "LastDataUplinkRate", sta->ul_rate);
	blobmsg_add_u32(bb, "UtilizationReceive", sta->dl_utilization);
	blobmsg_add_u32(bb, "UtilizationTransmit", sta->ul_utilization);
	blobmsg_add_u32(bb, "EstMACDataRateDownlink", sta->dl_est_thput);
	blobmsg_add_u32(bb, "EstMACDataRateUplink", sta->ul_est_thput);
	blobmsg_add_u32(bb, "SignalStrength", sta->rssi);
	blobmsg_add_u32(bb, "LastConnectTime", sta->conn_time);
	blobmsg_add_u64(bb, "BytesSent", sta->tx_bytes);
	blobmsg_add_u64(bb, "BytesReceived", sta->rx_bytes);
	blobmsg_add_u64(bb, "PacketsSent", sta->tx_pkts);
	blobmsg_add_u64(bb, "PacketsReceived", sta->rx_pkts);
	blobmsg_add_u64(bb, "ErrorsSent", sta->tx_errors);
	blobmsg_add_u64(bb, "ErrorsReceived", sta->rx_errors);
	blobmsg_add_u64(bb, "RetransCount", sta->rtx_pkts);

	blobmsg_add_u32(bb, "NumberOfMeasureReports", sta->num_bcn_reports);

	arr = blobmsg_open_array(bb, "MeasurementReport");
	for (i = 0; i < sta->num_bcn_reports; i++) {
		t = blobmsg_open_table(bb, "");
		 ; // list of base64
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);

	blobmsg_add_string(bb, "IPV4Address", "TODO IPv4Address");
	blobmsg_add_string(bb, "IPV6Address", "TODO IPv6Address");
	blobmsg_add_string(bb, "Hostname", sta->hostname);
	blobmsg_add_string(bb, "CellularDataPreference", "TODO");
	blobmsg_add_u32(bb, "ReAssociationDelay", 0); // TODO unsignedInt​(:65535)
	blobmsg_add_u32(bb, "TIDQueueSizesNumberOfEntries", 0); // TODO unsignedInt

	dump2_network_device_radio_bss_sta_multiapsta(p, net, dev, radio, bss, sta, bb);
	dump2_network_device_radio_bss_sta_wifi6cap(p, net, dev, radio, bss, sta, bb);

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.STA.{i}.TIDQueueSizes.{i}. */
	arr = blobmsg_open_array(bb, "TIDQueueSizes");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_u32(bb, "TID", 0); // TODO unsignedInt​(:255)
		blobmsg_add_u32(bb, "Size", 0); // TODO unsignedInt​(:255)
	}
	blobmsg_close_array(bb, arr); // TIDQueueSizes
}

static void dump2_network_device_radio_bss_sta_multiapsta(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		struct wifi_radio_element *radio, struct wifi_bss_element *bss,
		struct wifi_sta_element *sta, struct blob_buf *bb)
{
	void *arr;
	void *t, *t2;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.STA.{i}.MultiAPSTA. */
	t = blobmsg_open_table(bb, "MultiAPSTA");
	blobmsg_add_string(bb, "AssociationTime", "TODO dateTime");
	blobmsg_add_u32(bb, "Noise", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "SteeringHistoryNumberOfEntries", 0); // TODO unsignedInt

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.STA.{i}.MultiAPSTA.SteeringSummaryStats. */
	t2 = blobmsg_open_table(bb, "SteeringSummaryStats");
	blobmsg_add_u64(bb, "NoCandidateAPFailures", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BlacklistAttempts", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BlacklistSuccesses", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BlacklistFailures", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BTMAttempts", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BTMSuccesses", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BTMFailures", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BTMQueryResponses", 0); //TODO StatsCounter64
	blobmsg_add_u32(bb, "LastSteerTime", 0); // TODO unsignedInt
	blobmsg_close_table(bb, t2); // SteeringSummaryStats

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.STA.{i}.MultiAPSTA.SteeringHistory.{i}. */
	arr = blobmsg_open_array(bb, "SteeringHistory");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_string(bb, "Time", "TODO dateTime");
		blobmsg_add_string(bb, "APOrigin", "TODO");
		blobmsg_add_string(bb, "TriggerEvent", "TODO");
		blobmsg_add_string(bb, "SteeringApproach", "TODO");
		blobmsg_add_string(bb, "APDestination", "TODO");
		blobmsg_add_u32(bb, "SteeringDuration", 0); // TODO unsignedInt
	}
	blobmsg_close_array(bb, arr); // SteeringHistory

	blobmsg_close_table(bb, t); // MultiAPSTA
}

static void dump2_network_device_radio_bss_sta_wifi6cap(struct decollector_private *p,
		struct wifi_network *net, struct wifi_network_device *dev,
		struct wifi_radio_element *radio, struct wifi_bss_element *bss,
		struct wifi_sta_element *sta, struct blob_buf *bb)
{
	void *t;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.STA.{i}.WiFi6Capabilities. */
	t = blobmsg_open_table(bb, "WiFi6Capabilities");
	blobmsg_add_string(bb, "HE160", "TODO bool");
	blobmsg_add_string(bb, "HE8080", "TODO bool");
	blobmsg_add_string(bb, "MCSNSS", "TODO base64​(4:12)");
	blobmsg_add_string(bb, "SUBeamformer", "TODO bool");
	blobmsg_add_string(bb, "SUBeamformee", "TODO bool");
	blobmsg_add_string(bb, "MUBeamformer", "TODO bool");
	blobmsg_add_string(bb, "Beamformee80orLess", "TODO bool");
	blobmsg_add_string(bb, "BeamformeeAbove80", "TODO bool");
	blobmsg_add_string(bb, "ULMUMIMO", "TODO bool");
	blobmsg_add_string(bb, "ULOFDMA", "TODO bool");
	blobmsg_add_u32(bb, "MaxDLMUMIMO", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "MaxULMUMIMO", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "MaxDLOFDMA", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "MaxULOFDMA", 0); // TODO unsignedInt​(:255)
	blobmsg_add_string(bb, "RTS", "TODO bool");
	blobmsg_add_string(bb, "MURTS", "TODO bool");
	blobmsg_add_string(bb, "MultiBSSID", "TODO bool");
	blobmsg_add_string(bb, "MUEDCA", "TODO bool");
	blobmsg_add_string(bb, "TWTRequestor", "TODO bool");
	blobmsg_add_string(bb, "TWTResponder", "TODO bool");
	blobmsg_add_string(bb, "SpatialReuse", "TODO bool");
	blobmsg_add_string(bb, "AnticipatedChannelUsage", "TODO bool");
	blobmsg_close_table(bb, t); // WiFi6Capabilities
}
