/*
 * decollector_tlv.c
 * File contains TLVs handling needed for WiFi Data Elements.
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: saurabh.verma@iopsys.eu
 *	   anjan.chanda@iopsys.eu
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#include <easy/easy.h>

#include "timer.h"
#include <cmdu.h>
#include <1905_tlvs.h>
#include <map2.h>
#include <cmdu_ackq.h>

#include "wfa_delm.h"
#include "decollector.h"

int decollector_gen_aladdr(struct decollector_private *priv, struct cmdu_buff *cmdu)
{
	int ret;
	struct tlv *t;
	struct tlv_aladdr *data;

	t = cmdu_reserve_tlv(cmdu, 30);
	if (!t)
		return -1;

	t->type = TLV_TYPE_AL_MAC_ADDRESS_TYPE;
	t->len = sizeof(*data);
	data = (struct tlv_aladdr *)t->data;
	memcpy(data->macaddr, priv->ieee1905_macaddr, 6);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int decollector_gen_searched_role(struct decollector_private *priv,
		struct cmdu_buff *cmdu, uint8_t role)
{
	int ret;
	struct tlv *t;
	struct tlv_searched_role *data;

	t = cmdu_reserve_tlv(cmdu, 30);
	if (!t)
		return -1;

	t->type = TLV_TYPE_SEARCHED_ROLE;
	t->len = sizeof(*data);
	data = (struct tlv_searched_role *)t->data;
	data->role = role;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int decollector_gen_autoconfig_band(struct decollector_private *priv,
		struct cmdu_buff *cmdu, uint8_t band)
{
	int ret;
	struct tlv *t;
	struct tlv_autoconfig_band *data;

	t = cmdu_reserve_tlv(cmdu, 30);
	if (!t)
		return -1;

	t->type = TLV_TYPE_AUTOCONFIG_FREQ_BAND;
	t->len = sizeof(*data);
	data = (struct tlv_autoconfig_band *)t->data;
	data->band = band;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int decollector_gen_supported_service(struct decollector_private *priv,
		struct cmdu_buff *cmdu, uint8_t service)
{
	int i, ret;
	struct tlv *t;

	t = cmdu_reserve_tlv(cmdu, 30);
	if (!t)
		return -1;

	t->type = MAP_TLV_SUPPORTED_SERVICE;
	t->data[0] = 1;
	for (i = 0; i < t->data[0]; i++)
		t->data[1 + i] = service;

	t->len = 1 + t->data[0];

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int decollector_gen_searched_service(struct decollector_private *priv,
		struct cmdu_buff *cmdu, uint8_t service)
{
	int i, ret;
	struct tlv *t;

	t = cmdu_reserve_tlv(cmdu, 30);
	if (!t)
		return -1;

	t->type = MAP_TLV_SEARCHED_SERVICE;
	t->data[0] = 1;
	for (i = 0; i < t->data[0]; i++)
		t->data[1 + i] = service;

	t->len = 1 + t->data[0];

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int decollector_gen_map_profile(struct decollector_private *priv,
		struct cmdu_buff *cmdu, uint8_t profile)
{
	int ret;
	struct tlv *t;
	struct tlv_map_profile *data;

	t = cmdu_reserve_tlv(cmdu, 20);
	if (!t)
		return -1;

	t->type = MAP_TLV_MULTIAP_PROFILE;
	t->len = sizeof(*data);
	data = (struct tlv_map_profile *)t->data;
	data->profile = profile;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int decollector_gen_channel_pref(struct decollector_private *priv,
		struct cmdu_buff *cmdu, struct wifi_radio_element *radio)
{
	int i, ret;
	int offset = 0;
	struct tlv *t;
	struct tlv_channel_pref *data;
	struct wifi_opclass_supported_element *opclass;
	int reserve_len = 256;

	t = cmdu_reserve_tlv(cmdu, reserve_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_CHANNEL_PREFERENCE;
	data = (struct tlv_channel_pref *)t->data;

	memcpy(data->radio, radio->macaddr, 6);
	data->num_opclass = radio->num_supp_opclass;
	offset = sizeof(*data);
	for (i = 0; i < data->num_opclass; i++) {
		struct channel_pref_opclass *op =
			(struct channel_pref_opclass *)&t->data[offset];

		if ((offset + 3) > reserve_len)
			goto error;

		opclass = radio->supp_opclass + i;
		op->classid = opclass->id;
		op->chs.num_channel = 0;	/*include all channels */
		op->preference = 0x00;		/* TODO: Hardcoded value */
		offset += sizeof(*op);
	}

	/* update the tlv length */
	t->len = offset;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;

error:
	return -1;
}

int decollector_gen_txpower_limit(struct decollector_private *priv,
		struct cmdu_buff *cmdu, uint8_t *radio_id,
		uint8_t power_limit)
{
	int ret;
	struct tlv *t;
	struct tlv_txpower_limit *data;

	t = cmdu_reserve_tlv(cmdu, 20);
	if (!t)
		return -1;

	t->type = MAP_TLV_TRANSMIT_POWER_LIMIT;
	t->len = sizeof(*data);
	data = (struct tlv_txpower_limit *)t->data;
	memcpy(data->radio, radio_id, 6);
	data->limit = power_limit;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int decollector_gen_ap_radio_identifier(struct decollector_private *priv,
		struct cmdu_buff *cmdu, uint8_t *radio_id)
{
	int ret;
	struct tlv *t;
	struct tlv_ap_radio_identifier *data;

	t = cmdu_reserve_tlv(cmdu, 20);
	if (!t)
		return -1;

	t->type = MAP_TLV_AP_RADIO_IDENTIFIER;
	t->len = sizeof(*data);
	data = (struct tlv_ap_radio_identifier *)t->data;
	memcpy(data->radio, radio_id, 6);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int decollector_gen_ap_metric_query(struct decollector_private *priv,
		struct cmdu_buff *cmdu, int num_bss, uint8_t *bsslist)
{
	int i, ret;
	struct tlv *t;
	struct tlv_ap_metric_query *data;
	int reserve_len = 0;

	reserve_len = sizeof(*data) + (6 * num_bss);
	t = cmdu_reserve_tlv(cmdu, reserve_len + 4);
	if (!t)
		return -1;

	t->type = MAP_TLV_AP_METRIC_QUERY;
	t->len = reserve_len;
	data = (struct tlv_ap_metric_query *) t->data;

	data->num_bss = num_bss;
	for (i = 0; i < data->num_bss; i++)
		memcpy(data->bss[i].bssid, &bsslist[i * 6], 6);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int decollector_gen_channel_scan_request(struct decollector_private *priv,
		struct cmdu_buff *cmdu, struct wifi_network_device *dev)
{
	int i, j;
	int ret;
	int offset = 0;
	struct tlv *t;
	struct tlv_channel_scan_request *data;
	struct wifi_radio_element *radio;
	struct wifi_opclass_supported_element *opclass;
	int reserve_len = 1024;

	t = cmdu_reserve_tlv(cmdu, reserve_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_CHANNEL_SCAN_REQ;
	data = (struct tlv_channel_scan_request *)t->data;

	data->mode = SCAN_REQUEST_FRESH_SCAN;
	data->num_radio = dev->num_radios;
	offset = sizeof(*data);
	for (i = 0; i < data->num_radio; i++) {
		struct channel_scan_request_radio *r =
			(struct channel_scan_request_radio *)&t->data[offset];

		if ((offset + 6 + 1) > reserve_len)
			goto error;

		radio = dev->radiolist + i;
		memcpy(r->radio, radio->macaddr, 6);
		r->num_opclass = radio->num_supp_opclass;
		offset += sizeof(*r);

		for (j = 0; j < radio->num_supp_opclass; j++) {
			struct channel_scan_request_opclass *op =
				(struct channel_scan_request_opclass *)&t->data[offset];

			if ((offset + 2) > reserve_len)
				goto error;

			opclass = radio->supp_opclass + j;
			op->classid = opclass->id;
			/* Scan all the channels in the operating class */
			op->num_channel = 0;
			offset += sizeof(*op) + op->num_channel;
		}
	}

	/* update the tlv length */
	t->len = offset;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;

error:
	return -1;
}

static inline void put_duration_data(uint8_t *buf, uint32_t val)
{
	/* ignore val MSB data */
	buf[0] = (val >> 16) & 0x0ff;
	buf[1] = (val >> 8) & 0xff;
	buf[2] = val & 0xff;
}

#define AP_COLLECTION_INTERVAL	(10 * 1000)

int decollector_gen_client_info(struct decollector_private *priv,
               struct cmdu_buff *cmdu, uint8_t *sta,
               uint8_t *bssid)
{
	int ret;
	struct tlv *t;
	struct tlv_client_info *data;

	t = cmdu_reserve_tlv(cmdu, 256);
	if (!t)
		return -1;

	t->type = MAP_TLV_CLIENT_INFO;
	t->len = sizeof(*data);
	data = (struct tlv_client_info *)t->data;
	memcpy(data->bssid, bssid, 6);
	memcpy(data->macaddr, sta, 6);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

